"""
Utilities for working with
`Hugging Face Transformers <https://huggingface.co/docs/transformers>`_.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

from fiftyone.core.config import Config
import fiftyone.core.labels as fol
from fiftyone.core.models import Model, EmbeddingsMixin, PromptMixin
import fiftyone.core.utils as fou
from fiftyone.zoo.models import HasZooModel

torch = fou.lazy_import("torch")
transformers = fou.lazy_import("transformers")


logger = logging.getLogger(__name__)

DEFAULT_CLASSIFICATION_PATH = "google/vit-base-patch16-224"
DEFAULT_DETECTION_PATH = "hustvl/yolos-tiny"
DEFAULT_SEGMENTATION_PATH = "nvidia/segformer-b0-finetuned-ade-512-512"
DEFAULT_DEPTH_ESTIMATION_PATH = "Intel/dpt-hybrid-midas"
DEFAULT_ZERO_SHOT_CLASSIFICATION_PATH = "openai/clip-vit-large-patch14"
DEFAULT_ZERO_SHOT_DETECTION_PATH = "google/owlvit-base-patch32"


def convert_transformers_model(model, task=None):
    """Converts the given Hugging Face transformers model into a FiftyOne
    model.

    Args:
        model: a ``transformers`` model
        task (None): the task of the model. Supported values are
            ``"image-classification"``, ``"object-detection"``,
            ``"semantic-segmentation"``, and ``"depth-estimation"``.
            If not specified, the task is automatically inferred from the model

    Returns:
         a :class:`fiftyone.core.models.Model`

    Raises:
        ValueError: if the model could not be converted
    """
    model_type = get_model_type(model, task=task)

    if model_type == "zero-shot-image-classification":
        return _convert_zero_shot_transformer_for_image_classification(model)
    elif model_type == "zero-shot-object-detection":
        return _convert_zero_shot_transformer_for_object_detection(model)
    elif model_type == "image-classification":
        return _convert_transformer_for_image_classification(model)
    elif model_type == "object-detection":
        return _convert_transformer_for_object_detection(model)
    elif model_type == "semantic-segmentation":
        return _convert_transformer_for_semantic_segmentation(model)
    elif model_type == "depth-estimation":
        return _convert_transformer_for_depth_estimation(model)
    elif model_type == "base-model":
        return _convert_transformer_base_model(model)
    else:
        raise ValueError(
            "Unsupported model type; cannot convert %s to a FiftyOne model"
            % model
        )


def _convert_transformer_base_model(model):
    config = FiftyOneTransformerConfig({"model": model})
    return FiftyOneTransformer(config)


def _convert_transformer_for_image_classification(model):
    config = FiftyOneTransformerConfig({"model": model})
    return FiftyOneTransformerForImageClassification(config)


def _convert_transformer_for_object_detection(model):
    config = FiftyOneTransformerConfig({"model": model})
    return FiftyOneTransformerForObjectDetection(config)


def _convert_transformer_for_semantic_segmentation(model):
    config = FiftyOneTransformerConfig({"model": model})
    return FiftyOneTransformerForSemanticSegmentation(config)


def _convert_transformer_for_depth_estimation(model):
    config = FiftyOneTransformerConfig({"model": model})
    return FiftyOneTransformerForDepthEstimation(config)


def _convert_zero_shot_transformer_for_image_classification(model):
    config = FiftyOneZeroShotTransformerConfig({"model": model})
    return FiftyOneZeroShotTransformerForImageClassification(config)


def _convert_zero_shot_transformer_for_object_detection(model):
    config = FiftyOneZeroShotTransformerConfig({"model": model})
    return FiftyOneZeroShotTransformerForObjectDetection(config)


def get_model_type(model, task=None):
    """Returns the string model type for the given model.

    If the model is a zero-shot model, the task is appended to the model type.

    Args:
        model: a ``transformers.model`` model
        task (None): an optional task type

    Returns:
        the model type string
    """
    if task and isinstance(task, str):
        if "zero-shot" in task:
            if "detection" in task:
                return "zero-shot-object-detection"
            else:
                return "zero-shot-image-classification"

    supported_tasks = (
        "image-classification",
        "object-detection",
        "semantic-segmentation",
        "depth-estimation",
    )
    if task is not None and task not in supported_tasks:
        raise ValueError(
            f"Unknown task: {task}. Valid tasks are {supported_tasks}"
        )

    zs = _is_zero_shot_model(model)

    if zs and task is None:
        if _has_detection_model(model):
            task = "object-detection"
        else:
            task = "image-classification"
    elif not zs and task is None:
        if _is_transformer_for_image_classification(model):
            task = "image-classification"
        elif _is_transformer_for_object_detection(model):
            task = "object-detection"
        elif _is_transformer_for_semantic_segmentation(model):
            task = "semantic-segmentation"
        elif _is_transformer_for_depth_estimation(model):
            task = "depth-estimation"
        elif _is_transformer_base_model(model):
            task = "base-model"
        else:
            raise ValueError(f"Unknown model type: {model}")

    if zs:
        return "zero-shot-" + task

    return task


def to_classification(results, id2label):
    """Converts the Transformers classification results to FiftyOne format.

    Args:
        results: Transformers classification results
        id2label: Transformers ID to label mapping

    Returns:
        a single or list of :class:`fiftyone.core.labels.Classification`
    """
    logits = results.logits
    predicted_labels = logits.argmax(-1)

    logits = logits.cpu().numpy()
    label_classes = [id2label[int(i)] for i in predicted_labels]

    odds = np.exp(logits)
    confidences = np.max(odds, axis=1) / np.sum(odds, axis=1)

    if logits.shape[0] == 1:
        return fol.Classification(
            label=label_classes[0], confidence=confidences[0], logits=logits[0]
        )

    return [
        fol.Classification(
            label=label_classes[i],
            confidence=confidences[i],
            logits=logits[i],
        )
        for i in range(logits.shape[0])
    ]


def to_segmentation(results):
    """Converts the Transformers semantic segmentation results to FiftyOne
    format.

    Args:
        results: Transformers semantic segmentation results

    Returns:
        a single or list of :class:`fiftyone.core.labels.Segmentation`
    """
    masks = [r.cpu().numpy() for r in results]

    if len(results) == 1:
        return _create_segmentation(masks[0])

    return [_create_segmentation(masks[i]) for i in range(len(masks))]


def _create_segmentation(mask):
    return fol.Segmentation(mask=mask)


def to_heatmap(results):
    """Converts the Transformers depth estimation results to FiftyOne format.

    Args:
        results: Transformers depth estimation results

    Returns:
        a single or list of :class:`fiftyone.core.labels.Heatmap`
    """

    if len(results.shape) == 2:
        return _create_heatmap(results)

    if len(results) == 1:
        return _create_heatmap(results[0])

    return [_create_heatmap(results[i]) for i in range(len(results))]


def _create_heatmap(heatmap):
    ## normalize the heatmap
    heatmap /= np.max(heatmap)
    return fol.Heatmap(map=heatmap)


def to_detections(results, id2label, image_sizes):
    """Converts the Transformers detection results to FiftyOne format.

    Args:
        results: Transformers detection results
        id2label: Transformers ID to label mapping
        image_sizes: the list of image sizes

    Returns:
        a single or list of :class:`fiftyone.core.labels.Detections`
    """
    if isinstance(results, dict):
        return _to_detections(results, id2label, image_sizes[0])

    if len(results) == 1:
        return _to_detections(results[0], id2label, image_sizes[0])

    return [
        _to_detections(result, id2label, image_sizes[i])
        for i, result in enumerate(results)
    ]


def _to_detections(result, id2label, image_size):
    detections = []

    scores = result["scores"].cpu().numpy()
    labels = result["labels"].cpu().numpy()
    boxes = result["boxes"].cpu().numpy()
    for score, label, box in zip(scores, labels, boxes):
        box = [round(i, 2) for i in box.tolist()]
        box = _convert_bounding_box(box, image_size)
        detections.append(
            fol.Detection(
                label=id2label[label.item()],
                bounding_box=box,
                confidence=score.item(),
            )
        )

    return fol.Detections(detections=detections)


def _convert_bounding_box(box, image_shape):
    top_left_x, top_left_y, bottom_right_x, bottom_right_y = box

    width = bottom_right_x - top_left_x
    height = bottom_right_y - top_left_y

    img_width, img_height = image_shape

    return [
        top_left_x / img_width,
        top_left_y / img_height,
        width / img_width,
        height / img_height,
    ]


class FiftyOneTransformerConfig(Config, HasZooModel):
    """Configuration for a :class:`FiftyOneTransformer`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
    """

    def __init__(self, d):
        self.model = self.parse_raw(d, "model", default=None)
        self.name_or_path = self.parse_string(d, "name_or_path", default=None)
        if etau.is_str(self.model):
            self.name_or_path = self.model
            self.model = None


class FiftyOneZeroShotTransformerConfig(FiftyOneTransformerConfig):
    """Configuration for a :class:`FiftyOneZeroShotTransformer`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
        text_prompt: the text prompt to use, e.g., ``"A photo of"``
        classes (None): a list of custom classes for zero-shot prediction
    """

    def __init__(self, d):
        self.text_prompt = self.parse_string(d, "text_prompt", default=None)
        self.classes = self.parse_array(d, "classes", default=None)
        super().__init__(d)


class TransformerEmbeddingsMixin(EmbeddingsMixin):
    """Mixin for Transformers that can generate embeddings."""

    @property
    def has_embeddings(self):
        # If the model family supports classification or detection tasks, its
        # embeddings from last_hidden_layer are meaningful and properly sized
        smodel = str(type(self.model)).split(".")
        model_name = smodel[-1][:-2].split("For")[0].replace("Model", "")
        module_name = "transformers"

        classif_model_name = f"{model_name}ForImageClassification"
        detection_model_name = f"{model_name}ForObjectDetection"

        _dynamic_import = __import__(
            module_name, fromlist=[classif_model_name, detection_model_name]
        )

        return hasattr(_dynamic_import, classif_model_name) or hasattr(
            _dynamic_import, detection_model_name
        )

    def embed(self, arg):
        return self._embed(arg)[0]

    def embed_all(self, args):
        return self._embed(args)

    def _embed(self, args):
        inputs = self.image_processor(args, return_tensors="pt")
        with torch.no_grad():
            outputs = self.model.base_model(**inputs)

        return outputs.last_hidden_state[:, -1, :].cpu().numpy()


class ZeroShotTransformerEmbeddingsMixin(EmbeddingsMixin):
    """Mixin for Transformers that can generate embeddings."""

    @property
    def has_embeddings(self):
        return _has_text_and_image_features(self.model)

    def embed(self, arg):
        return self._embed(arg)[0]

    def embed_all(self, args):
        return self._embed(args)

    def _embed(self, args):
        inputs = self.processor(images=args, return_tensors="pt")
        with torch.no_grad():
            image_features = self.model.base_model.get_image_features(**inputs)

        return image_features.cpu().numpy()


class ZeroShotTransformerPromptMixin(PromptMixin):
    """Mixin for Transformers that can perform zero-shot prediction."""

    @property
    def can_embed_prompts(self):
        return _has_text_and_image_features(self.model)

    def embed_prompt(self, prompt):
        """Generates an embedding for the given text prompt.

        Args:
            prompt: a text string

        Returns:
            a numpy vector
        """
        return self.embed_prompts([prompt])[0]

    def embed_prompts(self, prompts):
        """Generates an embedding for the given text prompts.

        Args:
            prompts: an iterable of text strings

        Returns:
            a ``num_prompts x num_dims`` array of prompt embeddings
        """
        return self._embed_prompts(prompts).detach().cpu().numpy()

    def _embed_prompts(self, prompts):
        inputs = self.processor(text=prompts, return_tensors="pt")
        with torch.no_grad():
            text_features = self.model.base_model.get_text_features(**inputs)
        return text_features


class FiftyOneTransformer(TransformerEmbeddingsMixin, Model):
    """FiftyOne wrapper around a ``transformers`` model.

    Args:
        config: a `FiftyOneTransformerConfig`
    """

    def __init__(self, config):
        self.config = config
        self.model = self._load_model(config)
        self.image_processor = self._load_image_processor()

    @property
    def media_type(self):
        return "image"

    @property
    def ragged_batches(self):
        return False

    @property
    def transforms(self):
        return None

    @property
    def preprocess(self):
        return False

    def _load_image_processor(self):
        return _get_image_processor(self.model)

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        return transformers.AutoModel.from_pretrained(config.name_or_path)

    def predict(self, arg):
        raise NotImplementedError("Subclass must implement predict()")


class FiftyOneZeroShotTransformer(
    ZeroShotTransformerEmbeddingsMixin, ZeroShotTransformerPromptMixin, Model
):
    """FiftyOne wrapper around a ``transformers`` model.

    Args:
        config: a `FiftyOneZeroShotTransformerConfig`
    """

    def __init__(self, config):
        self.config = config
        self.classes = config.classes
        self.model = self._load_model(config)
        self.processor = self._load_processor()
        self._text_prompts = None

    @property
    def media_type(self):
        return "image"

    @property
    def ragged_batches(self):
        return False

    @property
    def transforms(self):
        return None

    @property
    def preprocess(self):
        return False

    def _load_processor(self):
        return _get_processor(self.model)

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        return transformers.AutoModel.from_pretrained(config.name_or_path)

    def _get_text_prompts(self):
        if self._text_prompts is not None:
            return self._text_prompts

        if self.classes is None and self.config.classes is None:
            return None

        if self.classes is None:
            self.classes = self.config.classes

        if self.config.text_prompt is None:
            self._text_prompts = self.classes
        else:
            self._text_prompts = [
                "%s %s" % (self.config.text_prompt, c) for c in self.classes
            ]

        return self._text_prompts

    def predict(self, arg):
        raise NotImplementedError("Subclass must implement predict()")


class FiftyOneZeroShotTransformerForImageClassificationConfig(
    FiftyOneZeroShotTransformerConfig
):
    """Configuration for a
    :class:`FiftyOneZeroShotTransformerForImageClassification`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
        text_prompt: the text prompt to use, e.g., ``"A photo of"``
        classes (None): a list of custom classes for zero-shot prediction
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_ZERO_SHOT_CLASSIFICATION_PATH


class FiftyOneZeroShotTransformerForImageClassification(
    FiftyOneZeroShotTransformer
):
    """FiftyOne wrapper around a ``transformers`` model for zero-shot image
    classification.

    Args:
        config: a `FiftyOneZeroShotTransformerConfig`
    """

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        model = transformers.AutoModel.from_pretrained(config.name_or_path)
        if _has_image_text_retrieval(model):
            model = _get_model_for_image_text_retrieval(
                model, config.name_or_path
            )

        return model

    def _predict_all_from_features(self, args):
        text_prompts = self._get_text_prompts()
        inputs = self.processor(
            images=args, text=text_prompts, return_tensors="pt", padding=True
        )

        with torch.no_grad():
            outputs = self.model(**inputs)

        logits_per_image = (
            outputs.logits_per_image.detach().cpu()
        )  # this is the image-text similarity score

        probs = logits_per_image.softmax(
            dim=1
        )  # we can take the softmax to get the label probabilities
        conf = probs.max(
            dim=1
        ).values.numpy()  # the confidence of the most likely label
        logits = (
            logits_per_image.numpy()
        )  # we can also take the logits and train a linear classifier on top
        arg_max = (
            probs.argmax(dim=1).numpy().astype(int)
        )  # the argmax of the label probabilities
        classifs = [
            fol.Classification(
                label=self.classes[arg_max[i]],
                confidence=conf[i],
                logits=logits[i],
            )
            for i in range(logits.shape[0])
        ]

        if len(classifs) == 1:
            classifs = classifs[0]

        return classifs

    def _predict_all_from_retrieval(self, args):
        return [self._predict_from_retrieval(arg) for arg in args]

    def _predict_from_retrieval(self, arg):
        text_prompts = self._get_text_prompts()
        logits = []

        with torch.no_grad():
            for text_prompt in text_prompts:
                inputs = self.processor(arg, text_prompt, return_tensors="pt")
                outputs = self.model(**inputs)
                logits.append(outputs.logits[0, :].item())

        logits = np.array(logits)
        probs = np.exp(logits) / np.sum(np.exp(logits))
        conf = probs.max()
        arg_max = int(probs.argmax())
        label = self.classes[arg_max]

        return fol.Classification(
            label=label,
            confidence=conf,
            logits=logits,
        )

    def predict_all(self, args):
        if _has_text_and_image_features(self.model):
            return self._predict_all_from_features(args)
        else:
            return self._predict_all_from_retrieval(args)

    def predict(self, arg):
        return self.predict_all(arg)


class FiftyOneTransformerForImageClassificationConfig(
    FiftyOneTransformerConfig
):
    """Configuration for a :class:`FiftyOneTransformerForImageClassification`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_CLASSIFICATION_PATH


class FiftyOneTransformerForImageClassification(FiftyOneTransformer):
    """FiftyOne wrapper around a ``transformers`` model for image
    classification.

    Args:
        config: a `FiftyOneTransformerConfig`
    """

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        return transformers.AutoModelForImageClassification.from_pretrained(
            config.name_or_path
        )

    def _predict(self, inputs):
        with torch.no_grad():
            results = self.model(**inputs)
        return to_classification(results, self.model.config.id2label)

    def predict(self, arg):
        inputs = self.image_processor(arg, return_tensors="pt")
        return self._predict(inputs)

    def predict_all(self, args):
        inputs = self.image_processor(args, return_tensors="pt")
        return self._predict(inputs)


class FiftyOneZeroShotTransformerForObjectDetectionConfig(
    FiftyOneZeroShotTransformerConfig
):
    """Configuration for a
    :class:`FiftyOneZeroShotTransformerForObjectDetection`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
        text_prompt: the text prompt to use, e.g., ``"A photo of"``
        classes (None): a list of custom classes for zero-shot prediction
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_ZERO_SHOT_DETECTION_PATH
        elif self.model is not None:
            self.name_or_path = self.model.name_or_path


class FiftyOneZeroShotTransformerForObjectDetection(
    FiftyOneZeroShotTransformer
):
    """FiftyOne wrapper around a ``transformers`` model for zero shot object
    detection.

    Args:
        config: a `FiftyOneZeroShotTransformerConfig`
    """

    def __init__(self, config):
        self.config = config
        self.classes = config.classes
        self.processor = self._load_processor(config)
        self.model = self._load_model(config)
        self._text_prompts = None

    def _load_processor(self, config):
        name_or_path = config.name_or_path
        if not name_or_path:
            if config.model is not None:
                name_or_path = config.model.name_or_path

        return transformers.AutoProcessor.from_pretrained(name_or_path)

    def _load_model(self, config):
        name_or_path = config.name_or_path
        if not name_or_path:
            if config.model is not None:
                name_or_path = config.model.name_or_path

        if config.model is not None:
            return config.model
        else:
            return _get_detector_from_processor(self.processor, name_or_path)

    def _process_inputs(self, args):
        text_prompts = self._get_text_prompts()
        inputs = self.processor(
            images=args, text=text_prompts, return_tensors="pt", padding=True
        )
        return inputs

    def _predict(self, inputs, target_sizes):
        with torch.no_grad():
            outputs = self.model(**inputs)

        results = self.processor.post_process_object_detection(
            outputs, target_sizes=target_sizes
        )
        image_shapes = [i[::-1] for i in target_sizes]

        id2label = {i: c for i, c in enumerate(self.classes)}
        return to_detections(results, id2label, image_shapes)

    def predict(self, arg):
        target_sizes = [arg.shape[:-1][::-1]]
        inputs = self._process_inputs(arg)
        return self._predict(inputs, target_sizes)


class FiftyOneTransformerForObjectDetectionConfig(FiftyOneTransformerConfig):
    """Configuration for a :class:`FiftyOneTransformerForObjectDetection`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_DETECTION_PATH


class FiftyOneTransformerForObjectDetection(FiftyOneTransformer):
    """FiftyOne wrapper around a ``transformers`` model for object detection.

    Args:
        config: a `FiftyOneTransformerConfig`
    """

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        return transformers.AutoModelForObjectDetection.from_pretrained(
            config.name_or_path
        )

    def _predict(self, inputs, target_sizes):
        with torch.no_grad():
            outputs = self.model(**inputs)

        results = self.image_processor.post_process_object_detection(
            outputs, target_sizes=target_sizes
        )
        image_shapes = [i[::-1] for i in target_sizes]
        return to_detections(results, self.model.config.id2label, image_shapes)

    def predict(self, arg):
        target_sizes = [arg.shape[:-1][::-1]]
        inputs = self.image_processor(arg, return_tensors="pt")
        return self._predict(inputs, target_sizes)

    def predict_all(self, args):
        target_sizes = [i.shape[:-1][::-1] for i in args]
        inputs = self.image_processor(args, return_tensors="pt")
        return self._predict(inputs, target_sizes)


class FiftyOneTransformerForSemanticSegmentationConfig(
    FiftyOneTransformerConfig
):
    """Configuration for a :class:`FiftyOneTransformerForSemanticSegmentation`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_SEGMENTATION_PATH


class FiftyOneTransformerForSemanticSegmentation(FiftyOneTransformer):
    """FiftyOne wrapper around a ``transformers`` model for semantic
    segmentation.

    Args:
        config: a `FiftyOneTransformerConfig`
    """

    def _load_model(self, config):
        if config.model is not None:
            model = config.model
        else:
            model = (
                transformers.AutoModelForSemanticSegmentation.from_pretrained(
                    config.name_or_path
                )
            )

        self.mask_targets = model.config.id2label
        return model

    def _predict(self, inputs, target_sizes):
        with torch.no_grad():
            outputs = self.model(**inputs)

        results = self.image_processor.post_process_semantic_segmentation(
            outputs, target_sizes=target_sizes
        )
        return to_segmentation(results)

    def predict(self, arg):
        target_sizes = [arg.shape[:-1][::-1]]
        inputs = self.image_processor(arg, return_tensors="pt")
        return self._predict(inputs, target_sizes)

    def predict_all(self, args):
        target_sizes = [i.shape[:-1][::-1] for i in args]
        inputs = self.image_processor(args, return_tensors="pt")
        return self._predict(inputs, target_sizes)


class FiftyOneTransformerForDepthEstimationConfig(FiftyOneTransformerConfig):
    """Configuration for a :class:`FiftyOneTransformerForDepthEstimation`.

    Args:
        model (None): a ``transformers`` model
        name_or_path (None): the name or path to a checkpoint file to load
    """

    def __init__(self, d):
        super().__init__(d)
        if self.model is None and self.name_or_path is None:
            self.name_or_path = DEFAULT_DEPTH_ESTIMATION_PATH


class FiftyOneTransformerForDepthEstimation(FiftyOneTransformer):
    """FiftyOne wrapper around a ``transformers`` model for depth estimation.

    Args:
        config: a `FiftyOneTransformerConfig`
    """

    def _load_model(self, config):
        if config.model is not None:
            return config.model

        return transformers.AutoModelForDepthEstimation.from_pretrained(
            config.name_or_path
        )

    def _predict(self, inputs, target_sizes):
        with torch.no_grad():
            outputs = self.model(**inputs)

        predicted_depth = outputs.predicted_depth
        prediction = torch.nn.functional.interpolate(
            predicted_depth.unsqueeze(1),
            size=target_sizes[0],
            mode="bicubic",
            align_corners=False,
        )
        prediction = prediction.squeeze(1).cpu().numpy()

        return to_heatmap(prediction)

    def predict(self, arg):
        target_sizes = [arg.shape[:-1][::-1]]
        inputs = self.image_processor(arg, return_tensors="pt")
        return self._predict(inputs, target_sizes)

    def predict_all(self, args):
        target_sizes = [i.shape[:-1][::-1] for i in args]
        inputs = self.image_processor(args, return_tensors="pt")
        return self._predict(inputs, target_sizes)


def _has_text_and_image_features(model):
    return hasattr(model.base_model, "get_image_features") and hasattr(
        model.base_model, "get_text_features"
    )


def _has_image_text_retrieval(model):
    module_name = "transformers"
    model_name = _get_model_type_string(model)
    itr_class_name = f"{model_name}ForImageAndTextRetrieval"
    return hasattr(
        __import__(module_name, fromlist=[itr_class_name]),
        itr_class_name,
    )


def _has_detection_model(model):
    module_name = "transformers"
    model_name = _get_base_model_name(model)
    detection_class_name = f"{model_name}ForObjectDetection"
    return hasattr(
        __import__(module_name, fromlist=[detection_class_name]),
        detection_class_name,
    )


def _is_zero_shot_model(model):
    if _has_text_and_image_features(model):
        return True

    if _has_image_text_retrieval(model):
        return True

    return False


def _get_model_type_string(model):
    return str(type(model)).split(".")[-1][:-2]


def _is_transformer_for_image_classification(model):
    return "ForImageClassification" in _get_model_type_string(model)


def _is_transformer_for_object_detection(model):
    return "ForObjectDetection" in _get_model_type_string(model)


def _is_transformer_for_semantic_segmentation(model):
    ms = _get_model_type_string(model)
    return "For" in ms and "Segmentation" in ms


def _is_transformer_for_depth_estimation(model):
    return "ForDepthEstimation" in _get_model_type_string(model)


def _is_transformer_base_model(model):
    model_type = _get_model_type_string(model)
    return "Model" in model_type and "For" not in model_type


def _get_base_model_name(model):
    return str(type(model)).split(".")[-1][:-2].split("For")[0]


def _get_model_for_image_text_retrieval(base_model, model_name_or_path):
    model_name = _get_base_model_name(base_model)
    module_name = "transformers"
    itr_class_name = f"{model_name}ForImageAndTextRetrieval"
    itr_class = getattr(
        __import__(module_name, fromlist=[itr_class_name]),
        itr_class_name,
    )
    return itr_class.from_pretrained(model_name_or_path)


def _get_image_processor_fallback(model):
    model_name = _get_base_model_name(model)
    module_name = "transformers"
    processor_class_name = f"{model_name}ImageProcessor"
    processor_class = getattr(
        __import__(module_name, fromlist=[processor_class_name]),
        processor_class_name,
    )
    return processor_class.from_pretrained(model.config.model_name_or_path)


def _get_image_processor(model):
    try:
        image_processor = transformers.AutoImageProcessor.from_pretrained(
            model.config._name_or_path
        )
    except:
        image_processor = _get_image_processor_fallback(model)

    return image_processor


def _get_processor(model):
    try:
        processor = transformers.AutoProcessor.from_pretrained(
            model.config._name_or_path
        )
    except:
        raise ValueError(
            "Could not find a processor for model %s"
            % model.config._name_or_path
        )

    return processor


def _get_detector_from_processor(processor, model_name_or_path):
    module_name = "transformers"
    processor_class_name = f"{processor.__class__.__name__}"
    detector_class_name = (
        f"{processor_class_name.split('Processor')[0]}ForObjectDetection"
    )
    detector_class = getattr(
        __import__(module_name, fromlist=[detector_class_name]),
        detector_class_name,
    )
    return detector_class.from_pretrained(model_name_or_path)



"""
Plotting utils.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import itertools
import warnings

import numpy as np
import sklearn.linear_model as skl
import sklearn.metrics as skm

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone.core.expressions as foe
import fiftyone.core.labels as fol


def parse_scatter_inputs(
    points,
    samples=None,
    ids=None,
    link_field=None,
    labels=None,
    sizes=None,
    edges=None,
    classes=None,
):
    points = np.asarray(points)
    num_dims = points.shape[1]

    if ids is None and samples is not None:
        if num_dims != 2:
            msg = "Interactive selection is only supported in 2D"
            warnings.warn(msg)
        else:
            ids = _get_ids(samples, link_field=link_field, ref=points)

    labels = _parse_values(
        labels,
        "labels",
        samples=samples,
        ids=ids,
        link_field=link_field,
        ref=points,
    )

    sizes = _parse_values(
        sizes,
        "sizes",
        samples=samples,
        ids=ids,
        link_field=link_field,
        ref=points,
    )

    if ids is not None:
        ids = np.asarray(ids)

    if sizes is not None:
        sizes = np.asarray(sizes)

    if edges is not None:
        edges = np.asarray(edges)

    if labels is None:
        return points, ids, None, sizes, edges, None, False

    labels = np.asarray(labels)

    if not etau.is_str(labels[0]):
        return points, ids, labels, sizes, edges, None, False

    if classes is None:
        classes = sorted(set(labels))
        return points, ids, labels, sizes, edges, classes, True

    found = np.array([l in classes for l in labels])
    if not np.all(found):
        points = points[found, :]
        labels = labels[found]

        if sizes is not None:
            sizes = sizes[found]

        if ids is not None:
            ids = ids[found]

        if edges is not None:
            i = set(np.nonzero(found)[0])
            edges = np.array([e for e in edges if e[0] in i and e[1] in i])

    return points, ids, labels, sizes, edges, classes, True


def parse_locations(locations, samples, ids=None):
    if locations is not None and not etau.is_str(locations):
        return np.asarray(locations)

    if samples is None:
        raise ValueError(
            "You must provide `samples` in order to extract `locations` from "
            "your dataset"
        )

    if locations is None:
        location_field = samples._get_geo_location_field()
    else:
        location_field = locations
        samples.validate_field_type(
            location_field, embedded_doc_type=fol.GeoLocation
        )

    path = location_field + ".point.coordinates"
    if ids is not None:
        locations = samples._get_values_by_id(path, ids)
    else:
        locations = samples.values(path)

    return np.asarray(locations)


def parse_lines_inputs(
    x=None,
    y=None,
    samples=None,
    ids=None,
    link_field=None,
    labels=None,
    sizes=None,
):
    if y is None:
        raise ValueError("You must provide 'y' values")

    raw_y_values = not (etau.is_str(y) or _is_expr(y))

    if raw_y_values:
        is_frames = y and etau.is_container(y[0])
    elif samples is not None and samples._contains_videos():
        is_frames = foe.is_frames_expr(y)
    else:
        is_frames = False

    if is_frames and link_field is None:
        link_field = "frames"

    if ids is None and samples is not None:
        ref = y if raw_y_values else None
        ids = _get_ids(
            samples, link_field=link_field, ref=ref, is_frames=is_frames
        )

    x = _parse_values(
        x,
        "x",
        samples=samples,
        ids=ids,
        link_field=link_field,
        is_frames=is_frames,
    )

    y = _parse_values(
        y,
        "y",
        samples=samples,
        ids=ids,
        link_field=link_field,
        is_frames=is_frames,
    )

    sizes = _parse_values(
        sizes,
        "sizes",
        samples=samples,
        ids=ids,
        link_field=link_field,
        ref=y,
        is_frames=is_frames,
    )

    if x is None:
        if is_frames:
            x = [np.arange(1, len(yi) + 1) for yi in y]
        else:
            x = np.arange(1, len(y) + 1)

    if is_frames and x and not etau.is_container(x[0]):
        x = [x] * len(y)

    if is_frames:
        if sizes is None:
            sizes = [None] * len(y)

        if ids is None:
            ids = [None] * len(y)

        if labels is None:
            labels = [str(i) for i in range(1, len(y) + 1)]
        else:
            labels = _parse_values(labels, "labels", samples=samples)
    else:
        x = [x]
        y = [y]
        ids = [ids]
        sizes = [sizes]
        labels = [labels]

    return x, y, ids, link_field, labels, sizes, is_frames


def best_fit_line(points, label=None):
    x = points[:, 0]
    y = points[:, 1]

    model = skl.LinearRegression()
    model.fit(x[:, np.newaxis], y)

    xline = np.array([x.min(), x.max()])
    yline = model.predict(xline[:, np.newaxis])

    if label is None:
        r2_score = skm.r2_score(x, y, sample_weight=None)
        label = "r^2: %0.3f" % r2_score

    return xline, yline, label


def _is_expr(arg):
    return isinstance(arg, (foe.ViewExpression, dict))


def _parse_values(
    values,
    parameter,
    samples=None,
    ids=None,
    link_field=None,
    ref=None,
    is_frames=False,
):
    if values is None:
        return None

    if etau.is_str(values) or isinstance(values, foe.ViewExpression):
        if samples is None:
            raise ValueError(
                "You must provide `samples` in order to extract field values "
                "for the `%s` parameter" % parameter
            )

        if ids is not None and not is_frames:
            values = samples._get_values_by_id(
                values, ids, link_field=link_field
            )
        else:
            values = samples.values(values)

    if is_frames:
        values = [_unwind_values(v) for v in values]
    else:
        values = _unwind_values(values)

    if ref is not None:
        _validate_values(values, ref, parameter, is_frames=is_frames)

    return values


def _get_ids(samples, link_field=None, ref=None, is_frames=False):
    if link_field is None:
        ids = samples.values("id")
        ptype = "sample"
    elif link_field == "frames":
        ids = samples.values("frames.id")
        ptype = "frame"
    else:
        _, id_path = samples._get_label_field_path(link_field, "id")
        ids = samples.values(id_path)
        ptype = "label"

    if is_frames:
        ids = [_unwind_values(_ids) for _ids in ids]
    else:
        ids = _unwind_values(ids)

    if ref is not None:
        values_type = "%s IDs" % ptype
        _validate_values(ids, ref, values_type, is_frames=is_frames)

    return ids


def _unwind_values(values):
    if values is None:
        return None

    while any(etau.is_container(v) for v in values):
        values = list(itertools.chain.from_iterable(v for v in values if v))

    return np.array(values)


def _validate_values(values, ref, values_type, is_frames=False):
    if not is_frames:
        if len(values) != len(ref):
            raise ValueError(
                "Inconsistent number of %s (%d != %d). You may have missing "
                "data/labels that you need to omit from your view"
                % (values_type, len(values), len(ref))
            )

        return

    if len(values) != len(ref):
        raise ValueError(
            "Inconsistent number of %s traces (%d != %d). You may have "
            "missing data/labels that you need to omit from your view"
            % (values_type, len(values), len(ref))
        )

    for idx, (_values, _ref) in enumerate(zip(values, ref), 1):
        if len(_values) != len(_ref):
            raise ValueError(
                "Inconsistent number of %s (%d != %d) in trace %d/%d. You may "
                "have missing data/labels that you need to omit from your view"
                % (values_type, len(_values), len(_ref), idx, len(values))
            )


def load_button_icon(name):
    return load_icon(_BUTTON_ICONS[name])


def load_icon(icon_bytes):
    alpha = etas.deserialize_numpy_array(icon_bytes)
    h, w = alpha.shape
    img = np.zeros((h, w, 4), dtype=alpha.dtype)
    img[:, :, -1] = alpha
    return img


def serialize_icon(img):
    if img.shape[2] < 4 or np.any(img[:, :, :-1] > 0):
        raise ValueError("Image must be alpha-only")

    alpha = img[:, :, -1]
    return etas.serialize_numpy_array(alpha)


def pad_icon(img, pad, fill=0):
    h, w, c = img.shape[:3]
    img_pad = np.full((h + 2 * pad, w + 2 * pad, c), fill, dtype=img.dtype)
    img_pad[pad : (pad + h), pad : (pad + w), :] = img
    return img_pad


# Original images from https://www.iconfinder.com
_BUTTON_ICONS = {
    "map": "eJztk79LG2Ech9+jcgdddFKaLDcYTiWNuCTQVeymyWCGTiXEJEdbGklMl7aT4FKERAeJk/4Dpd1bx3avm0bQrFl00Io/+nr33q/3vfu+74VS6PI+w4V88+QZPiHbS8XFwgsFvUPvjZVKs9wwnunGh9ackdaNar2x1ii9fVlvrFTs+/PSm2bFujfN0mrFej+Vy6b1XHY6rX/U/5LHSCL5LxQ3ng7lZQoTMUbqBGN8+iS2lPqK8eWCyBjbvMM2v9vi2tjmja0d8Y0Rc4A9rjuJITyuMn+Iabg1yuOU7AFCgDXGA0vuALG1kCccAF+R5xlcowfltIIB7rtl8qpWqdpWEvBegS1qgINZtOw6aq0fqsEeZ4BeXkGUw9ZmuJ6HUvUHuGhq9oV26BoWeA7r/gA7486FdbRwjePZnLvKN//vHHbYGt9DyDGOlxTEdzTTq4k9cvihUhfo99FMctwTe+SwjITOsJ5syZZsyZZsyZZsyRZB3ScnU4tpqT/jWkr+GDv02RrP4zuZ7ziAqXG8c15rvHuPGfo1NcZbh1ta6wJH8GugN6gqUEsp9Dzj5tPM1rVfO3NqoDeKENCaPfC//GXSOifDNY4Xbb3e/eMZv+bdDxIdpkZey4AXbvkMao+Cj5gaeV6BHtiKDEDXRF60FR0ArkEe24IGgGo8L2jBA7i1dlATeehINEC4div2Fi6tAVLCEql9tlK9OG+ikIkt2aQ3ikN5Esk/5wFTASSi",
    "sync": "eJztll9IU1Ecx28u7zWVCp1BS2tPbtP1UISDIIIeon8zJ/2hBwnJifRHZTezKImQ2V5FEgwpIoIKzTR76yGSWAZFT6EvZS77o840N/8s9XbVc86uu+fcc67tKfZ92u73y+ee39nvd86aCo87i06t4S5xVy1lbvGMx7LbbLlWs8NiN1vKqzwXPaWVp6s8Ze7F5/tLz4tu+blYUVrtlr9bHQV2s6PAZjfXmVepVC6hhBKKkcEhtvYMBCML4UDv/Zo9PDLyusOhLis7iC9un5BWKPTkpLCMWjLGWWFGb1DCKNiwSTa7l790MpHSfGEcaVFhbzoHzEkWlCtAIi0q4ISf6KSU21okpaio7A+sKCrL+oUZRWNZh5Xheb/XlZ/BJ623FTX0LuhkZStXNVhrUno5dd/1sFIUezVSnqyyxXF2Vks09yADF9j8nJXlQqmIW2XKM6hjv9JRi4YPqlETsSRNlg+t6oDa7FajNFhZqAZ1gRyHG1Ayy4u2HeeG1CjybPPwkBnG/oJdalYHkVUMI+VY2zoeixrLJbLaQWRQ1aIA1jm5osAOMmrtbxCqJUaY5QCoeRM9S5MIWP5/R3GtgOWNA6sHsFxxYA0AVn4cWLBTsY2qUxHA4unR+LN2vpgOP8vDWnprtC1NwYQN5+ndezBybTgP9kQxG0qYBgOOM2GvNrCx9oH4DM70APMNG6sJxPtxJprtLSwoHl7v93CuAd4zl1lYJfAYO4G124AbYOiwpI/wREzH+uiiraCz3DDbivf5UeCPZtJQWXC3FuyERD182UMaC26H9JSUyERXIKXKKpibIw8Jumz/HNZCOedgrpkcSkV/5KY0YIemYGpoo8YbCyW0srOkTFkEhY5orZ5rRjnpkREX2HA3mmjURHHC+2g0eE6ItQ0l36L+a8L9jmT6HA1LQ1dylF7mhT6F2ZdFQXFc7k9FXpp/e/PYdqMgGG2u668iSiewlYqSYcqVEdW/jQEll/mOjuqlF7gs4RYN1ajj4nMOaJG+HmUnyUqtnySRZn1pulCyMm6M4EhjPqYjPFbJhY9/rQSF2kt0rwnJsKu65eWnkdmZH/3+O5V7160alFBC/63+Asl9Dog=",
    "disconnect": "eJztmG9IE2Ecx28ribQ/qGQUledGTHFSYWpYRilYUYqCRoYh1i6tLG2rREosaC8Mk2gQlI2cK6NQ8nUJ0ZtSiBaCEigpkdpfnTqkctu16e+5u9nzPHc4e7fvu/t+v8/n7rk9z90xS25hTl6RirnEXNYZONMJoy6d1V25uE2nZ3UnK40XjKXnSiqNBs7vZ5dWmDifbyovreJ8x/FpqXo2LTVBz9ayC1Q4E1JIIf1vxZjs9vNrFgV1dIr3afLwIqAOeflZebKDRqk/8qAPqmBZKbygLcGyCkVWXrCsIyKrYAHDUzumXG0aMkvb7ppxHFSEOubxD/wWR2Jpv8/+rEUKUAfccyOfkljtc0d/smRRm50wcorEcsHhWJwMKuwdGumUY/HdS+msq8LIFhLrmWDUUFGaX6g3up7ESvYgY3ojjfUY1b4mMCQWwwmOjYKKR6ecTkYWbq3WI8etwXP8uoNKnGDh9tCSN8i6RUSFT0DluehtF1lJgpk4A9bPZSQWugZPouipBxCqT9K8h8wcEssOhTapuQ/uoTtTYmrRjb1PYo1AITPALZyd+Xh+gNkJ1SECKhbyUXWgH11htZ6JCvTK0CTX4Vm5EDeTrlsiLWLtx+cmiMsUsJgfUD6Ljxsh3qWE9RrK9fjYBjGrhNUG5SZ8/ATiaCWsB1B+iI9bIY5RwkJrkbC970JM2bCiOqBswcc3IN6thPUWymZ8fBpigwKUaoJezoLYqoClR2s1A5/HQDyogFUJXW8kodALhR3yrC6oOkgFCxSIDxJBwhQbSI09UPi9QY71CLHSSQ3VEDRaZFAp6FHYT+7UotPRvxXC0OLiq8mlyEnojFD30U2EGltJaZlRqzuCXBLftXW0M674jGovV5E6JW7UGVxOYzEFwjkd+E8idZ3Q4HOpKIZpEZrOYkzMvhBRssswok8sv5q/2aKvT4tpD32Gfmm+iHW+69RaIQjba3VJouFNsiiG2erkpeq115YXH6+63ekKsMeT5El+2Cgvq2FlKN80e+VQPbEKUb6Pp2Y6qkn+tkuU/4lMGpJbV/MVcc2JJ43VLeD/pNU1A/+S+quJe0tGOxvfe0WO19FAfPQpUlQGZ7bYbBazIYP0mggppJAWSX8Bi5hqdQ==",
}



"""
FiftyOne Server view.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from typing import List, Optional

from bson import ObjectId
import strawberry as gql

import fiftyone.core.collections as foc
import fiftyone.core.dataset as fod
from fiftyone.core.expressions import ViewField as F, VALUE
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.stages as fosg
import fiftyone.core.utils as fou
import fiftyone.core.view as fov

from fiftyone.server.aggregations import GroupElementFilter, SampleFilter
from fiftyone.server.scalars import BSONArray, JSON


_LABEL_TAGS = "_label_tags"


@gql.input
class ExtendedViewForm:
    filters: Optional[JSON] = None
    mixed: Optional[bool] = None
    sample_ids: Optional[List[str]] = None
    slice: Optional[str] = None


async def load_view(
    dataset_name: str,
    serialized_view: BSONArray,
    form: ExtendedViewForm,
    view_name: Optional[str] = None,
) -> foc.SampleCollection:
    def run() -> foc.SampleCollection:
        dataset = fod.load_dataset(dataset_name)
        if view_name:
            view = dataset.load_saved_view(view_name)
            if serialized_view:
                for stage in serialized_view:
                    view.add_stage(fosg.ViewStage._from_dict(stage))
        else:
            view = get_view(
                dataset_name,
                stages=serialized_view,
                filters=form.filters,
                sample_filter=SampleFilter(
                    group=GroupElementFilter(slice=form.slice)
                ),
            )

            if form.sample_ids:
                view = fov.make_optimized_select_view(view, form.sample_ids)

            if form.mixed and view.media_type == fom.GROUP:
                view = view.select_group_slices(_force_mixed=True)
                view = get_extended_view(view, form.filters)

            return view

    return await fou.run_sync_task(run)


def get_view(
    dataset,
    view_name=None,
    stages=None,
    filters=None,
    pagination_data=False,
    extended_stages=None,
    sample_filter=None,
    reload=True,
):
    """Gets the view defined by the given request parameters.

    Args:
        dataset: the dataset name or :class:`fiftyone.core.dataset.Dataset`
            instance
        view_name (None): the name of a saved view to load
        stages (None): an optional list of serialized
            :class:`fiftyone.core.stages.ViewStage` instances
        filters (None): an optional ``dict`` of App defined filters
        pagination_data (False): whether process samples as pagination data
            - excludes all :class:`fiftyone.core.fields.DictField` values
            - filters label fields
        extended_stages (None): extended view stages
        sample_filter (None): an optional
            :class:`fiftyone.server.filters.SampleFilter`
        reload (None): whether to reload the dataset

    Returns:
        a :class:`fiftyone.core.view.DatasetView`
    """
    if isinstance(dataset, str):
        dataset = fod.load_dataset(dataset)

    if reload:
        dataset.reload()

    if view_name is not None:
        return dataset.load_saved_view(view_name)

    if stages:
        view = fov.DatasetView._build(dataset, stages)
    else:
        view = dataset.view()

    if sample_filter is not None:
        if sample_filter.group:
            if sample_filter.group.slice:
                view.group_slice = sample_filter.group.slice

            if sample_filter.group.id:
                view = fov.make_optimized_select_view(
                    view, sample_filter.group.id, groups=True
                )

            if sample_filter.group.slices:
                view = view.select_group_slices(
                    sample_filter.group.slices,
                    _force_mixed=True,
                )

        elif sample_filter.id:
            view = fov.make_optimized_select_view(view, sample_filter.id)

    if filters or extended_stages or pagination_data:
        view = get_extended_view(
            view,
            filters,
            pagination_data=pagination_data,
            extended_stages=extended_stages,
        )

    return view


def get_extended_view(
    view,
    filters=None,
    extended_stages=None,
    pagination_data=False,
):
    """Create an extended view with the provided filters.

    Args:
        view: a :class:`fiftyone.core.collections.SampleCollection`
        filters: an optional ``dict`` of App defined filters
        extended_stages (None): extended view stages
        pagination_data (False): filters label data

    Returns:
        a :class:`fiftyone.core.view.DatasetView`
    """
    label_tags = None

    if extended_stages:
        # extend view with similarity search, etc. first
        view = extend_view(view, extended_stages)

    if pagination_data:
        # omit all dict field values for performance, not needed by grid
        view = _project_pagination_paths(view)

    if filters:
        if "tags" in filters:
            tags = filters.get("tags")
            exclude = tags["exclude"]
            tags = tags["values"]

            if tags and not exclude:
                view = view.match_tags(tags)

            if tags and exclude:
                view = view.match_tags(tags, bool=False)

        label_tags = filters.get(_LABEL_TAGS, None)
        if label_tags:
            view = _match_label_tags(view, label_tags)

        stages = _make_filter_stages(view, filters)

        for stage in stages:
            view = view.add_stage(stage)

    if pagination_data:
        view = _add_labels_tags_counts(view)

    return view


def extend_view(view, extended_stages):
    """Adds the given extended stages to the view.

    Args:
        view: a :class:`fiftyone.core.collections.SampleCollection`
        extended_stages: an extended stages dict

    Returns:
        a :class:`fiftyone.core.view.DatasetView`
    """
    for _cls, d in extended_stages.items():
        kwargs = [[k, v] for k, v in d.items()]
        stage = fosg.ViewStage._from_dict({"_cls": _cls, "kwargs": kwargs})
        view = view.add_stage(stage)

    return view


def _add_labels_tags_counts(view):
    view = view.set_field(_LABEL_TAGS, [], _allow_missing=True)

    for path, field in foc._iter_label_fields(view):
        if isinstance(field, fof.ListField) or (
            isinstance(field, fof.EmbeddedDocumentField)
            and issubclass(field.document_type, fol._HasLabelList)
        ):
            if path.startswith(view._FRAMES_PREFIX):
                add_tags = _add_frame_labels_tags
            else:
                add_tags = _add_labels_tags
        else:
            if path.startswith(view._FRAMES_PREFIX):
                add_tags = _add_frame_label_tags
            else:
                add_tags = _add_label_tags

        view = add_tags(path, field, view)

    view = _count_list_items(_LABEL_TAGS, view)

    return view


def _project_pagination_paths(view: foc.SampleCollection):
    schema = view.get_field_schema(flat=True)
    frame_schema = view.get_frame_field_schema(flat=True)
    if frame_schema:
        schema.update(
            **{f"frames.{path}": field for path, field in frame_schema.items()}
        )

    excluded = [
        path
        for path, field in schema.items()
        if isinstance(field, fof.DictField)
    ]

    return view.select_fields(
        [
            path
            for path in schema
            if all(not path.startswith(exclude) for exclude in excluded)
        ]
    )


def _make_filter_stages(
    view,
    filters,
):
    stages = []
    queries = []
    for path, label_path, field, args in _iter_paths(view, filters):
        is_matching = args.get("isMatching", True)
        path_field = view.get_field(path)
        is_label_field = _is_label(field)
        if (
            is_label_field
            and issubclass(field.document_type, (fol.Keypoint, fol.Keypoints))
            and isinstance(path_field, (fof.KeypointsField, fof.ListField))
        ):
            continue

        if args.get("exclude") and not is_matching:
            continue

        queries.append(_make_query(path, path_field, args))

    if queries:
        stages.append(fosg.Match({"$and": queries}))

    for path, label_path, label_field, args in _iter_paths(
        view, filters, labels=True
    ):
        is_matching = args.get("isMatching", True)
        field = view.get_field(path)
        if issubclass(
            label_field.document_type, (fol.Keypoint, fol.Keypoints)
        ) and isinstance(field, fof.ListField):
            expr = _make_keypoint_list_filter(args, view, path, field)
            if expr is not None:
                stages.append(
                    fosg.FilterKeypoints(
                        label_path,
                        only_matches=True,
                        **expr,
                    )
                )

        elif not is_matching:
            key = field.db_field if field.db_field else field.name
            expr = _make_scalar_expression(F(key), args, field, is_label=True)
            if expr is not None:
                stages.append(
                    fosg.FilterLabels(
                        label_path,
                        expr,
                        only_matches=not args.get("exclude", False),
                    )
                )

    return stages


def _iter_paths(view, filters, labels=False):
    for path in sorted(filters):
        if path == "tags" or path.startswith("_"):
            continue

        if "." in path:
            parent_path = ".".join(path.split(".")[:-1])
        else:
            parent_path = path

        parent_field = view.get_field(parent_path)
        if isinstance(parent_field, fof.ListField) and isinstance(
            parent_field.field, fof.EmbeddedDocumentField
        ):
            if issubclass(parent_field.field.document_type, fol.Label):
                parent_path = ".".join(parent_path.split(".")[:-1])
                parent_field = view.get_field(parent_path)

        if labels and not _is_label(parent_field):
            continue

        yield path, parent_path, parent_field, filters[path]


def _is_support(field):
    if isinstance(field, fof.FrameSupportField):
        return True

    if isinstance(field, fof.EmbeddedDocumentField):
        if field.document_type in (
            fol.TemporalDetection,
            fol.TemporalDetections,
        ):
            return True

    return False


def _is_datetime(field):
    return isinstance(field, (fof.DateField, fof.DateTimeField))


def _is_label(field):
    return isinstance(field, fof.EmbeddedDocumentField) and issubclass(
        field.document_type, fol.Label
    )


def _make_query(path: str, field: fof.Field, args):
    keys = path.split(".")
    path = ".".join(keys[:-1] + [field.db_field or field.name])
    if isinstance(field, fof.ListField) and field.field:
        field = field.field

    if isinstance(
        field,
        (fof.DateField, fof.DateTimeField, fof.FloatField, fof.IntField),
    ):
        return _make_range_query(path, field, args)

    values = args.get("values", None)
    if isinstance(field, fof.ObjectIdField):
        values = list(map(lambda v: ObjectId(v), args["values"]))

    if isinstance(field, (fof.ObjectIdField, fof.StringField)):
        return {path: {"$nin" if args["exclude"] else "$in": values}}

    if isinstance(field, fof.BooleanField):
        true, false = args["true"], args["false"]
        if true and false:
            return {
                path: {"$nin" if args["exclude"] else "$in": [True, False]}
            }

        if not true and false:
            return {path: {"$ne" if args["exclude"] else "$eq": False}}

        if true and not false:
            return {path: {"$ne" if args["exclude"] else "$eq": True}}

        if not true and not false:
            return {
                path: {"$in" if args["exclude"] else "$nin": [True, False]}
            }

    raise ValueError("unexpected filter")


_NONFINITES = {
    "inf": float("inf"),
    "nan": float("nan"),
    "ninf": float("-inf"),
    "none": None,
}


def _make_range_query(path: str, field: fof.Field, args):
    range_ = args.get("range", None)
    if range_:
        mn, mx = range_
        if isinstance(field, (fof.DateField, fof.DateTimeField)):
            mn, mx = [fou.timestamp_to_datetime(d) for d in range_]
    else:
        mn, mx = None, None

    exclude = args.get("exclude", False)
    if not exclude and range_:
        return {
            "$and": [
                {path: {"$gte": mn}},
                {path: {"$lte": mx}},
            ]
        }

    if range_:
        return {
            "$or": [
                {path: {"$lt": mn}},
                {path: {"$gt": mx}},
            ]
        }

    if exclude:
        return {
            "$and": [
                {path: {"$eq": v}}
                for k, v in _NONFINITES.items()
                if k in args and not args[k]
            ]
        }

    return {
        "$and": [
            {path: {"$ne": v}}
            for k, v in _NONFINITES.items()
            if k in args and not args[k]
        ]
    }


def _make_scalar_expression(f, args, field, list_field=False, is_label=False):
    expr = None
    if _is_support(field):
        mn, mx = args["range"]
        expr = (f[0] >= mn) & (f[1] <= mx)
    elif isinstance(field, fof.ListField):
        expr = f.filter(
            _make_scalar_expression(F(), args, field.field, list_field=True)
        ).length()
        return expr == 0 if args["exclude"] else expr > 0
    elif isinstance(field, fof.BooleanField):
        true, false = args["true"], args["false"]
        if true and false:
            expr = f.is_in([True, False])

        if not true and false:
            expr = f == False

        if true and not false:
            expr = f == True

        if not true and not false:
            expr = (f != True) & (f != False)
    elif _is_datetime(field):
        mn, mx = args["range"]
        p = fou.timestamp_to_datetime
        expr = (f >= p(mn)) & (f <= p(mx))
    elif isinstance(field, (fof.FloatField, fof.IntField)):
        mn, mx = args["range"]
        expr = (f >= mn) & (f <= mx)
    else:
        values = args["values"]
        if not values:
            return None

        if isinstance(field, fof.ObjectIdField):
            f = f.to_string()

        none = any(map(lambda v: v is None, values))
        values = filter(lambda v: v is not None, values)
        expr = f.is_in(values)
        exclude = args["exclude"]

        if exclude:
            # pylint: disable=invalid-unary-operand-type
            expr = ~expr

        if none and not is_label and not list_field:
            if exclude:
                expr &= f.exists()
            else:
                expr |= ~(f.exists())

        return expr

    return _apply_others(expr, f, args, is_label)


def _make_keypoint_list_filter(args, view, path, field):
    name = path.split(".", 1)[0]

    if path.endswith(".points"):
        dataset = view._dataset
        if name in dataset.skeletons:
            skeleton = dataset.skeletons[name]
        else:
            skeleton = dataset.default_skeleton

        values = args.get("values", [])
        if args["exclude"]:
            values = list(set(skeleton.labels).difference(values))

        return {"labels": values}

    if isinstance(field.field, fof.BooleanField):
        true, false = args["true"], args["false"]
        f = F(name)
        if true and false:
            expr = f.is_in([True, False])

        if not true and false:
            expr = f == False

        if true and not false:
            expr = f == True

        if not true and not false:
            expr = (f != True) & (f != False)

        return {"filter": expr}

    if isinstance(field.field, (fof.FloatField, fof.IntField)):
        f = F(name)
        mn, mx = args["range"]
        expr = (f >= mn) & (f <= mx)
        if args["exclude"]:
            expr = ~expr

        return {"filter": expr}

    if isinstance(field.field, (fof.StringField)):
        f = F(name)
        if "values" in args:
            val = args["values"]
            expr = f.is_in(val)
            if args["exclude"]:
                expr = ~expr

            return {"filter": expr}

    raise ValueError(f"Filtering {field} keypoint fields is not supported")


def _apply_others(expr, f, args, is_label):
    is_matching = args.get("isMatching", False)
    nonfinites = {
        "nan": float("nan"),
        "ninf": -float("inf"),
        "inf": float("inf"),
    }
    include = []
    for k, v in nonfinites.items():
        if k in args and args[k]:
            include.append(v)

    if expr is None:
        expr = f.is_in(include)
    else:
        expr |= f.is_in(include)

    if "none" in args:
        expr = _apply_none(expr, f, args["none"])

    if (
        "exclude" in args
        and args["exclude"]
        and not (is_matching and is_label)
    ):
        expr = ~expr

    return expr


def _apply_none(expr, f, none):
    if not none:
        if expr is not None:
            expr &= f.exists()
        else:
            expr = f.exists()
    elif expr is not None:
        expr |= ~(f.exists())

    return expr


def _add_frame_labels_tags(path, field, view):
    path = path[len("frames.") :]
    items = path
    if isinstance(field, fof.ListField):
        field = field.field

    if issubclass(field.document_type, fol._HasLabelList):
        items = "%s.%s" % (path, field.document_type._LABEL_LIST_FIELD)

    reduce = F(items).reduce(VALUE.extend(F("tags")), [])
    view = view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend(
            F("frames").reduce(
                VALUE.extend(F(items).exists().if_else(reduce, [])),
                [],
            )
        ),
        _allow_missing=True,
    )
    return view


def _add_frame_label_tags(path, field, view):
    path = path[len("frames.") :]
    tags = "%s.tags" % path
    view = view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend(
            F("frames").reduce(
                VALUE.extend((F(tags) != None).if_else(F(tags), [])), []
            )
        ),
        _allow_missing=True,
    )
    return view


def _add_labels_tags(path, field, view):
    items = path
    if isinstance(field, fof.ListField):
        field = field.field

    if issubclass(field.document_type, fol._HasLabelList):
        items = "%s.%s" % (path, field.document_type._LABEL_LIST_FIELD)

    view = view.set_field(
        _LABEL_TAGS,
        F(path)
        .exists()
        .if_else(
            F(_LABEL_TAGS).extend(
                F(items).reduce(VALUE.extend(F("tags")), [])
            ),
            F(_LABEL_TAGS),
        ),
        _allow_missing=True,
    )
    return view


def _add_label_tags(path, field, view):
    tags = "%s.tags" % path
    return view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend((F(tags) != None).if_else(F(tags), [])),
        _allow_missing=True,
    )


def _count_list_items(path, view):
    function = (
        "function(items) {"
        "let counts = {};"
        "items && items.forEach((i) => {"
        "counts[i] = 1 + (counts[i] || 0);"
        "});"
        "return counts;"
        "}"
    )

    return view.set_field(
        path, F(path)._function(function), _allow_missing=True
    )


def _match_label_tags(view: foc.SampleCollection, label_tags):
    label_paths = [
        f"{path}.{field.document_type._LABEL_LIST_FIELD}"
        if isinstance(field, fof.EmbeddedDocumentField)
        and issubclass(field.document_type, fol._HasLabelList)
        else path
        for path, field in foc._iter_label_fields(view)
    ]
    values = label_tags["values"]
    exclude = label_tags["exclude"]
    matching = label_tags["isMatching"]
    expr = lambda exclude, values: {"$nin" if exclude else "$in": values}

    if not exclude or matching:
        view = view.mongo(
            [
                {
                    "$match": {
                        "$or": [
                            {f"{path}.tags": expr(exclude, values)}
                            for path in label_paths
                        ]
                    }
                }
            ]
        )

    if not matching and exclude:
        view = view.exclude_labels(
            tags=label_tags["values"],
            omit_empty=False,
            fields=view._get_label_fields(),
        )
    elif not matching:
        view = view.select_labels(
            tags=label_tags["values"],
            fields=view._get_label_fields(),
        )

    return view



"""
FiftyOne Server /frames route

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from starlette.endpoints import HTTPEndpoint
from starlette.responses import JSONResponse
from starlette.requests import Request

from fiftyone.core.expressions import ViewField as F
import fiftyone.core.json as foj
import fiftyone.core.odm as foo
import fiftyone.core.view as fov

from fiftyone.server.decorators import route
import fiftyone.core.media as fom
import fiftyone.server.view as fosv


class Frames(HTTPEndpoint):
    @route
    async def post(self, request: Request, data: dict):
        start_frame = int(data.get("frameNumber", 1))
        frame_count = int(data.get("frameCount", 1))
        num_frames = int(data.get("numFrames"))
        extended = data.get("extended", None)
        dataset = data.get("dataset")
        stages = data.get("view")
        sample_id = data.get("sampleId")
        group_slice = data.get("slice", None)

        view = fosv.get_view(dataset, stages=stages, extended_stages=extended)
        view = fov.make_optimized_select_view(view, sample_id)

        if view.media_type == fom.GROUP and group_slice is not None:
            view.group_slice = group_slice

        end_frame = min(num_frames + start_frame, frame_count)
        support = None if stages else [start_frame, end_frame]
        if not support:
            view = view.set_field(
                "frames",
                F("frames").filter(
                    (F("frame_number") >= start_frame)
                    & (F("frame_number") <= end_frame)
                ),
            )

        frames = await foo.aggregate(
            foo.get_async_db_conn()[view._dataset._sample_collection_name],
            view._pipeline(frames_only=True, support=support),
        ).to_list(end_frame - start_frame + 1)

        return JSONResponse(
            {
                "frames": foj.stringify(frames),
                "range": [start_frame, end_frame],
            }
        )



"""
Evaluation runs framework.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from fiftyone.core.runs import (
    BaseRun,
    BaseRunInfo,
    BaseRunConfig,
    BaseRunResults,
)
from fiftyone.core.odm import patch_evaluations


class EvaluationInfo(BaseRunInfo):
    """Information about an evaluation that has been run on a dataset.

    Args:
        key: the evaluation key
        timestamp (None): the UTC ``datetime`` when the evaluation was run
        config (None): the :class:`EvaluationMethodConfig` for the evaluation
    """

    @classmethod
    def config_cls(cls):
        return EvaluationMethodConfig


class EvaluationMethodConfig(BaseRunConfig):
    """Base class for configuring :class:`EvaluationMethod` instances.

    Args:
        **kwargs: any leftover keyword arguments after subclasses have done
            their parsing
    """

    @property
    def type(self):
        return "evaluation"

    @property
    def method(self):
        return None


class EvaluationMethod(BaseRun):
    """Base class for evaluation methods.

    Subclasses will typically declare an interface method that handles
    performing evaluation on an image, video, or entire collection.

    Args:
        config: an :class:`EvaluationMethodConfig`
    """

    @classmethod
    def run_info_cls(cls):
        return EvaluationInfo

    @classmethod
    def _runs_field(cls):
        return "evaluations"

    @classmethod
    def _run_str(cls):
        return "evaluation"

    @classmethod
    def _results_cache_field(cls):
        return "_evaluation_cache"

    @classmethod
    def _patch_function(cls):
        return patch_evaluations


class EvaluationResults(BaseRunResults):
    """Base class for evaluation results."""

    pass



"""
CSV utilities.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import csv
import os

import eta.core.utils as etau

import fiftyone.core.utils as fou
import fiftyone.core.sample as fos
import fiftyone.utils.data as foud


class CSVDatasetImporter(
    foud.GenericSampleDatasetImporter, foud.ImportPathsMixin
):
    """A flexible CSV importer that represents slice(s) of field values of a
    dataset as columns of a CSV file.

    See :ref:`this page <CSVDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``data_path``
            and/or ``labels_path`` must be provided
        data_path (None): an optional parameter that enables explicit control
            over the location of the media. Can be any of the following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``dataset_dir`` where the media files reside
            -   an absolute directory path where the media files reside. In
                this case, the ``dataset_dir`` has no effect on the location of
                the data
            -   a filename like ``"data.json"`` specifying the filename of the
                JSON data manifest file in ``dataset_dir``
            -   an absolute filepath specifying the location of the JSON data
                manifest. In this case, ``dataset_dir`` has no effect on the
                location of the data
            -   a dict mapping filenames to absolute filepaths

            If None, this parameter will default to whichever of ``data/`` or
            ``data.json`` exists in the dataset directory
        labels_path (None): an optional parameter that enables explicit control
            over the location of the labels. Can be any of the following:

            -   a filename like ``"labels.csv"`` specifying the location of
                the labels in ``dataset_dir``
            -   an absolute filepath to the labels. In this case,
                ``dataset_dir`` has no effect on the location of the labels

            If None, the parameter will default to ``labels.csv``
        media_field ("filepath"): the name of the column containing the media
            path for each sample. The media paths in this column may be:

            -   filenames or relative paths to media files in ``data_path``
            -   absolute media paths, in which case ``data_path`` has no effect
        fields (None): an optional parameter that specifies the columns to read
            and parse from the CSV file. Can be any of the following:

            -   an iterable of column names to parse as strings
            -   a dict mapping column names to functions that parse the column
                values into the appropriate type. Any keys with ``None`` values
                in this case are directly loaded as strings

            If not provided, all columns are parsed as strings
        skip_missing_media (False): whether to skip (True) or raise an error
            (False) when rows with no ``media_field`` are encountered
        include_all_data (False): whether to generate samples for all media in
            the data directory (True) rather than only creating samples for
            media with CSV rows (False)
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir=None,
        data_path=None,
        labels_path=None,
        media_field="filepath",
        fields=None,
        skip_missing_media=False,
        include_all_data=False,
        shuffle=False,
        seed=None,
        max_samples=None,
    ):
        if dataset_dir is None and data_path is None and labels_path is None:
            raise ValueError(
                "At least one of `dataset_dir`, `data_path`, and "
                "`labels_path` must be provided"
            )

        data_path = self._parse_data_path(
            dataset_dir=dataset_dir,
            data_path=data_path,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=labels_path,
            default="labels.csv",
        )

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.data_path = data_path
        self.labels_path = labels_path
        self.media_field = media_field
        self.fields = fields
        self.skip_missing_media = skip_missing_media
        self.include_all_data = include_all_data

        self._media_paths_map = None
        self._rows_map = None
        self._fields = None
        self._filepaths = None
        self._iter_filepaths = None
        self._num_samples = None

    def __iter__(self):
        self._iter_filepaths = iter(self._filepaths)
        return self

    def __len__(self):
        return self._num_samples

    def __next__(self):
        filepath = next(self._iter_filepaths)

        row = self._rows_map.get(filepath, None)
        if row is None:
            return fos.Sample(filepath=filepath)

        kwargs = {}
        if isinstance(self._fields, dict):
            for key, parser in self._fields.items():
                value = row.get(key, "")
                if value:
                    kwargs[key] = parser(value)
        elif self._fields is not None:
            for key in self._fields:
                value = row.get(key, "")
                if value:
                    kwargs[key] = value
        else:
            for key, value in row.items():
                if value:
                    kwargs[key] = value

        return fos.Sample(filepath=filepath, **kwargs)

    @property
    def has_sample_field_schema(self):
        return False

    @property
    def has_dataset_info(self):
        return False

    def setup(self):
        media_paths_map = self._load_data_map(self.data_path, recursive=True)

        rows_map = {}
        fields = _parse_import_fields(self.fields, self.media_field)

        media_field = self.media_field
        if self.labels_path is not None and os.path.isfile(self.labels_path):
            with open(self.labels_path, "r") as f:
                reader = csv.DictReader(f)
                if media_field not in reader.fieldnames:
                    raise ValueError(
                        "Media field '%s' not found in '%s'"
                        % (media_field, self.labels_path)
                    )

                for row in reader:
                    filename = row.pop(media_field, None)
                    if filename is None or os.path.isabs(filename):
                        filepath = filename
                    else:
                        filepath = media_paths_map.get(filename, None)

                    if filepath is None:
                        if self.skip_missing_media:
                            continue
                        else:
                            raise ValueError(
                                "Found row with no '%s' value" % media_field
                            )

                    rows_map[filepath] = row

        filepaths = set(rows_map.keys())

        if self.include_all_data:
            filepaths.update(media_paths_map.values())

        filepaths = self._preprocess_list(sorted(filepaths))

        self._media_paths_map = media_paths_map
        self._rows_map = rows_map
        self._fields = fields
        self._filepaths = filepaths
        self._num_samples = len(filepaths)


def _parse_import_fields(fields, media_field):
    if isinstance(fields, dict):
        fields.pop(media_field, None)
        fields = {k: v or str for k, v in fields.items()}
    elif fields is not None:
        if etau.is_str(fields):
            fields = [fields]

        fields = [f for f in fields if f != media_field]

    return fields


class CSVDatasetExporter(foud.BatchDatasetExporter, foud.ExportPathsMixin):
    """A flexible CSV exporter that represents slice(s) of field values of a
    dataset as columns of a CSV file.

    See :ref:`this page <CSVDataset-export>` for exporting datasets of this
    type.

    Args:
        export_dir (None): the directory to write the export. This has no
            effect if ``data_path`` and ``labels_path`` are absolute paths
        data_path (None): an optional parameter that enables explicit control
            over the location of the exported media. Can be any of the
            following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``export_dir`` in which to export the media
            -   an absolute directory path in which to export the media. In
                this case, the ``export_dir`` has no effect on the location of
                the data
            -   a JSON filename like ``"data.json"`` specifying the filename of
                the manifest file in ``export_dir`` generated when
                ``export_media`` is ``"manifest"``
            -   an absolute filepath specifying the location to write the JSON
                manifest file when ``export_media`` is ``"manifest"``. In this
                case, ``export_dir`` has no effect on the location of the data

            If None, the default value of this parameter will be chosen based
            on the value of the ``export_media`` parameter
        labels_path (None): an optional parameter that enables explicit control
            over the location of the exported labels. Can be any of the
            following:

            -   a filename like ``"labels.csv"`` specifying the location in
                ``export_dir`` in which to export the labels
            -   an absolute filepath to which to export the labels. In this
                case, the ``export_dir`` has no effect on the location of the
                labels

            If None, the labels will be exported into ``export_dir`` using the
            default filename
        export_media (None): controls how to export the raw media. The
            supported values are:

            -   ``True``: copy all media files into the output directory
            -   ``False``: don't export media
            -   ``"move"``: move all media files into the output directory
            -   ``"symlink"``: create symlinks to the media files in the output
                directory
            -   ``"manifest"``: create a ``data.json`` in the output directory
                that maps UUIDs used in the labels files to the filepaths of
                the source media, rather than exporting the actual media

            If None, the default value of this parameter will be chosen based
            on the value of the ``data_path`` parameter
        rel_dir (None): an optional relative directory to strip from each input
            filepath to generate a unique identifier for each media. When
            exporting media, this identifier is joined with ``data_path`` to
            generate an output path for each exported media. This argument
            allows for populating nested subdirectories that match the shape of
            the input paths. The path is converted to an absolute path (if
            necessary) via :func:`fiftyone.core.storage.normalize_path`
        abs_paths (False): whether to store absolute paths to the media in the
            exported labels
        media_field ("filepath"): the name of the field containing the media to
            export for each sample
        fields (None): an optional argument specifying the fields or
            ``embedding.field.names`` to include as columns in the exported
            CSV. Can be:

            -   a field or iterable of fields
            -   a dict mapping field names to column names

            By default, only the ``media_field`` is exported
    """

    def __init__(
        self,
        export_dir=None,
        data_path=None,
        labels_path=None,
        export_media=None,
        rel_dir=None,
        abs_paths=False,
        media_field="filepath",
        fields=None,
    ):
        data_path, export_media = self._parse_data_path(
            export_dir=export_dir,
            data_path=data_path,
            export_media=export_media,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            export_dir=export_dir,
            labels_path=labels_path,
            default="labels.csv",
        )

        super().__init__(export_dir=export_dir)

        self.data_path = data_path
        self.labels_path = labels_path
        self.export_media = export_media
        self.rel_dir = rel_dir
        self.abs_paths = abs_paths
        self.media_field = media_field
        self.fields = fields

        self._media_exporter = None
        self._f = None
        self._csv_writer = None
        self._paths = None
        self._media_idx = None
        self._include_media = None
        self._needs_metadata = None

    def setup(self):
        self._media_exporter = foud.MediaExporter(
            self.export_media,
            export_path=self.data_path,
            rel_dir=self.rel_dir,
        )
        self._media_exporter.setup()

        (
            paths,
            header,
            media_idx,
            include_media,
            needs_metadata,
        ) = _parse_export_fields(
            self.fields, self.media_field, self.export_media
        )

        etau.ensure_basedir(self.labels_path)
        f = open(self.labels_path, "w", newline="")

        # QUOTE_MINIMAL ensures that list fields are handled properly
        csv_writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(header)

        self._f = f
        self._csv_writer = csv_writer
        self._paths = paths
        self._media_idx = media_idx
        self._include_media = include_media
        self._needs_metadata = needs_metadata

    def export_samples(self, sample_collection, progress=None):
        if self._needs_metadata:
            sample_collection.compute_metadata()

        idx = self._media_idx
        with fou.ProgressBar(total=sample_collection, progress=progress) as pb:
            for data in pb(zip(*sample_collection.values(self._paths))):
                data = [_parse_value(d) for d in data]

                if idx is not None:
                    media_path = data[idx]
                    out_media_path, uuid = self._media_exporter.export(
                        media_path
                    )
                    if self._include_media:
                        data[idx] = out_media_path if self.abs_paths else uuid
                    else:
                        data.pop(idx)

                self._csv_writer.writerow(data)

    def close(self, *args):
        self._f.close()
        self._media_exporter.close()


def _parse_export_fields(fields, media_field, export_media):
    if not fields:
        fields = media_field

    if isinstance(fields, dict):
        paths, header = zip(*fields.items())
    elif etau.is_str(fields):
        paths = [fields]
        header = [fields]
    else:
        paths = list(fields)
        header = list(fields)

    try:
        media_idx = paths.index(media_field)
        include_media = True
    except ValueError:
        include_media = False
        if export_media:
            paths.append(media_field)
            media_idx = len(paths) - 1
        else:
            media_idx = None

    needs_metadata = any(p.startswith("metadata.") for p in paths)

    return paths, header, media_idx, include_media, needs_metadata


def _parse_value(value):
    if value is None:
        return ""

    if not etau.is_container(value):
        return str(value)

    # Render lists as "list,of,values"
    return ",".join(str(v) for v in value)



"""
Session server-sent events client.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from dataclasses import asdict, dataclass
import logging
from retrying import retry
from threading import Thread, Event as ThreadEvent
import time
import typing as t

from bson import json_util
import requests
import sseclient
from uuid import uuid4

import fiftyone.constants as foc
import fiftyone.core.state as fos

from fiftyone.core.session.events import (
    Event,
    EventType,
    ListenPayload,
    dict_factory,
)


logger = logging.getLogger(__name__)


@retry(wait_fixed=500, stop_max_delay=10000)
def _ping(url: str) -> None:
    requests.get(url)


@dataclass
class Client:
    address: str
    auto: bool
    desktop: bool
    port: int
    remote: bool
    start_time: float

    def __post_init__(self) -> None:
        self._subscription = str(uuid4())
        self._connected = True
        self._closed = ThreadEvent()
        self._closed.set()
        self._listeners: t.Dict[str, t.Set[t.Callable]] = defaultdict(set)

    @property
    def origin(self) -> str:
        """The origin of the server"""
        return f"http://{self.address}:{self.port}"

    @property
    def is_open(self) -> str:
        """Whether the client is connected"""
        return not self._closed.is_set()

    def open(self, state: fos.StateDescription) -> None:
        """Open the client connection

        Arg:
            state: the initial state description
        """
        if not self._closed.is_set():
            raise RuntimeError("Client is already running")

        def run_client() -> None:
            def subscribe() -> None:
                response = requests.post(
                    f"{self.origin}/events",
                    stream=True,
                    headers={
                        "Accept": "text/event-stream",
                        "Content-type": "application/json",
                    },
                    data=json_util.dumps(
                        asdict(
                            ListenPayload(
                                events=[
                                    "capture_notebook_cell",
                                    "close_session",
                                    "reactivate_notebook_cell",
                                    "refresh",
                                    "reload_session",
                                    "select_labels",
                                    "select_samples",
                                    "set_color_scheme",
                                    "set_dataset_color_scheme",
                                    "set_group_slice",
                                    "set_spaces",
                                    "state_update",
                                ],
                                initializer=state,
                                subscription=self._subscription,
                            ),
                            dict_factory=dict_factory,
                        )
                    ),
                )
                source = sseclient.SSEClient(response)
                for message in source.events():
                    event = Event.from_data(message.event, message.data)
                    self._dispatch_event(event)

            while True:
                try:
                    _ping(f"{self.origin}/fiftyone")
                    self._connected = True
                    subscribe()
                except Exception as e:
                    if logger.level == logging.DEBUG or foc.DEV_INSTALL:
                        raise e

                if self._closed.is_set():
                    break

                self._connected = False
                print(
                    "\r\nCould not connect session, trying again "
                    "in 10 seconds\r\n"
                )
                time.sleep(10)

        self._thread = Thread(target=run_client, daemon=True)
        self._closed.clear()
        self._thread.start()

    def close(self):
        """Close the client connection"""
        self._closed.set()
        self._thread.join(timeout=0)
        self._thread = None

    def send_event(self, event: EventType) -> None:
        """Sends an event to the server

        Args:
            event: the event
        """
        if self._closed.is_set():
            return

        if not self._connected:
            raise RuntimeError("Client is not connected")

        self._post_event(event)
        self._dispatch_event(event)

    def add_event_listener(
        self, event_name: str, listener: t.Callable
    ) -> None:
        """Adds an event listener callback for the provided event name. Events
        sent from client and from the server connection will be dispatched to
        the listener

        Args:
            event_name: the event name
            listener: the listener callback
        """
        self._listeners[event_name].add(listener)

    def remove_event_listener(
        self, event_name: str, listener: t.Callable
    ) -> None:
        """Removes an event listener callback for the provided event name if
        it has been registered

        Args:
            event_name: the event name
            listener: the listener callback
        """
        self._listeners[event_name].discard(listener)

    def _dispatch_event(self, event: EventType) -> None:
        for listener in self._listeners[event.get_event_name()]:
            listener(event)

    def _post_event(self, event: Event) -> None:
        if self._closed.is_set():
            return

        response = requests.post(
            f"{self.origin}/event",
            headers={"Content-type": "application/json"},
            data=json_util.dumps(
                {
                    "event": event.get_event_name(),
                    "data": asdict(event, dict_factory=dict_factory),
                    "subscription": self._subscription,
                }
            ),
        )

        if response.status_code != 200:
            raise RuntimeError(
                f"Failed to post event `{event.get_event_name()}` to {self.origin}/event"
            )



"""
FiftyOne delegated operations.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import asyncio
import logging
import traceback

from fiftyone.factory.repo_factory import RepositoryFactory
from fiftyone.factory import DelegatedOperationPagingParams
from fiftyone.operators.executor import (
    prepare_operator_executor,
    do_execute_operator,
    ExecutionResult,
    ExecutionRunState,
)


logger = logging.getLogger(__name__)


class DelegatedOperationService(object):
    """Service for executing delegated operations."""

    def __init__(self, repo=None):
        if repo is None:
            repo = RepositoryFactory.delegated_operation_repo()

        self._repo = repo

    def queue_operation(
        self, operator, label=None, delegation_target=None, context=None
    ):
        """Queues the given delegated operation for execution.

        Args:
            operator: the operator name
            delegation_target (None): an optional delegation target
            label (None): an optional label for the operation (will default to
                the operator if not supplied)
            context (None): an
                :class:`fiftyone.operators.executor.ExecutionContext`

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.queue_operation(
            operator=operator,
            label=label if label else operator,
            delegation_target=delegation_target,
            context=context,
        )

    def set_progress(self, doc_id, progress):
        """Sets the progress of the given delegated operation.

        Args:
            doc_id: the ID of the delegated operation
            progress: the
                :class:`fiftyone.operators.executor.ExecutionProgress` of the
                operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.update_progress(_id=doc_id, progress=progress)

    def set_running(self, doc_id, progress=None, run_link=None):
        """Sets the given delegated operation to running state.

        Args:
            doc_id: the ID of the delegated operation
            progress (None): an optional
                :class:`fiftyone.operators.executor.ExecutionProgress` of the
                operation
            run_link (None): an optional link to orchestrator-specific
                information about the operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.update_run_state(
            _id=doc_id,
            run_state=ExecutionRunState.RUNNING,
            run_link=run_link,
            progress=progress,
        )

    def set_completed(
        self,
        doc_id,
        result=None,
        progress=None,
        run_link=None,
    ):
        """Sets the given delegated operation to completed state.

        Args:
            doc_id: the ID of the delegated operation
            result (None): the
                :class:`fiftyone.operators.executor.ExecutionResult` of the
                operation
            progress (None): an optional
                :class:`fiftyone.operators.executor.ExecutionProgress` of the
                operation
            run_link (None): an optional link to orchestrator-specific
                information about the operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.update_run_state(
            _id=doc_id,
            run_state=ExecutionRunState.COMPLETED,
            result=result,
            progress=progress,
            run_link=run_link,
        )

    def set_failed(
        self,
        doc_id,
        result=None,
        progress=None,
        run_link=None,
    ):
        """Sets the given delegated operation to failed state.

        Args:
            doc_id: the ID of the delegated operation
            result (None): the
                :class:`fiftyone.operators.executor.ExecutionResult` of the
                operation
            progress (None): an optional
                :class:`fiftyone.operators.executor.ExecutionProgress` of the
                operation
            run_link (None): an optional link to orchestrator-specific
                information about the operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.update_run_state(
            _id=doc_id,
            run_state=ExecutionRunState.FAILED,
            result=result,
            run_link=run_link,
            progress=progress,
        )

    def set_pinned(self, doc_id, pinned=True):
        """Sets the pinned flag for the given delegated operation.

        Args:
            doc_id: the ID of the delegated operation
            pinned (True): the boolean pinned flag

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.set_pinned(_id=doc_id, pinned=pinned)

    def set_label(self, doc_id, label):
        """Sets the pinned flag for the given delegated operation.

        Args:
            doc_id: the ID of the delegated operation
            label: the label to set

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.set_label(_id=doc_id, label=label)

    def delete_operation(self, doc_id):
        """Deletes the given delegated operation.

        Args:
            doc_id: the ID of the delegated operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.delete_operation(_id=doc_id)

    def delete_for_dataset(self, dataset_id):
        """Deletes all delegated operations associated with the given dataset.

        Args:
            dataset_id: the ID of the dataset
        """
        return self._repo.delete_for_dataset(dataset_id=dataset_id)

    def rerun_operation(self, doc_id):
        """Reruns the specified delegated operation.

        Args:
            doc_id: the ID of the delegated operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        doc = self._repo.get(_id=doc_id)
        return self._repo.queue_operation(**doc.__dict__)

    def get_queued_operations(self, operator=None, dataset_name=None):
        """Returns all queued delegated operations.

        Args:
            operator (None): the optional name of the operator to return all
                the queued delegated operations for
            dataset_name (None): the optional name of the dataset to return all
                the queued delegated operations for

        Returns:
            a list of :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.get_queued_operations(
            operator=operator, dataset_name=dataset_name
        )

    def get(self, doc_id):
        """Returns the delegated operation with the given ID.

        Args:
            doc_id: the ID of the delegated operation

        Returns:
            a :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.get(_id=doc_id)

    def list_operations(
        self,
        operator=None,
        dataset_name=None,
        dataset_id=None,
        run_state=None,
        delegation_target=None,
        paging=None,
        search=None,
        **kwargs,
    ):
        """Lists the delegated operations matching the given criteria.

        Args:
            operator (None): the optional name of the operator to return all
                the delegated operations for
            dataset_name (None): the optional name of the dataset to return all
                the delegated operations for
            dataset_id (None): the optional id of the dataset to return all the
                delegated operations for
            run_state (None): the optional run state of the delegated
                operations to return
            delegation_target (None): the optional delegation target of the
                delegated operations to return
            paging (None): optional
                :class:`fiftyone.factory.DelegatedOperationPagingParams`
            search (None): optional search term dict

        Returns:
            a list of :class:`fiftyone.factory.repos.DelegatedOperationDocument`
        """
        return self._repo.list_operations(
            operator=operator,
            dataset_name=dataset_name,
            dataset_id=dataset_id,
            run_state=run_state,
            delegation_target=delegation_target,
            paging=paging,
            search=search,
            **kwargs,
        )

    def execute_queued_operations(
        self,
        operator=None,
        delegation_target=None,
        dataset_name=None,
        limit=None,
        log=False,
        **kwargs,
    ):
        """Executes queued delegated operations matching the given criteria.

        Args:
            operator (None): the optional name of the operator to execute all
                the queued delegated operations for
            delegation_target (None): the optional delegation target of the
                delegated operations to execute
            dataset_name (None): the optional name of the dataset to execute
                all the queued delegated operations for
            limit (None): the optional limit of the number of delegated
                operations to execute
            log (False): the optional boolean flag to log the execution of the
                delegated operations
        """
        if limit is not None:
            paging = DelegatedOperationPagingParams(limit=limit)
        else:
            paging = None

        queued_ops = self.list_operations(
            operator=operator,
            dataset_name=dataset_name,
            delegation_target=delegation_target,
            run_state=ExecutionRunState.QUEUED,
            paging=paging,
            **kwargs,
        )

        for op in queued_ops:
            self.execute_operation(operation=op, log=log)

    def count(self, filters=None, search=None):
        """Counts the delegated operations matching the given criteria.

        Args:
            filters (None): a filter dict
            search (None): a search term dict

        Returns:
            the number of matching operations
        """
        return self._repo.count(filters=filters, search=search)

    def execute_operation(self, operation, log=False, run_link=None):
        """Executes the given delegated operation.

        Args:
            operation: the
                :class:`fiftyone.factory.repos.DelegatedOperationDocument`
            log (False): the optional boolean flag to log the execution of the
                delegated operations
            run_link (None): an optional link to orchestrator-specific
                information about the operation
        """
        try:
            self.set_running(doc_id=operation.id, run_link=run_link)
            if log:
                logger.info(
                    "\nRunning operation %s (%s)",
                    operation.id,
                    operation.operator,
                )

            result = asyncio.run(self._execute_operator(operation))

            self.set_completed(doc_id=operation.id, result=result)
            if log:
                logger.info("Operation %s complete", operation.id)
        except:
            result = ExecutionResult(error=traceback.format_exc())

            self.set_failed(doc_id=operation.id, result=result)
            if log:
                logger.info(
                    "Operation %s failed\n%s", operation.id, result.error
                )

    async def _execute_operator(self, doc):
        operator_uri = doc.operator
        context = doc.context
        context.request_params["run_doc"] = doc.id

        if not context.request_params.get("dataset_id", None):
            # Pass the dataset_id so that the execution context can load the
            # dataset by id in case a dataset has been renamed since the task
            # was initially queued. However, don't overwrite it if it exists.
            context.request_params["dataset_id"] = doc.dataset_id

        prepared = await prepare_operator_executor(
            operator_uri=operator_uri,
            request_params=context.request_params,
            delegated_operation_id=doc.id,
            set_progress=self.set_progress,
        )

        if isinstance(prepared, ExecutionResult):
            raise prepared.to_exception()

        operator, _, ctx = prepared
        return await do_execute_operator(operator, ctx, exhaust=True)



"""
FiftyOne Server routes

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from .aggregate import Aggregate
from .embeddings import EmbeddingsRoutes
from .event import Event
from .events import Events
from .fiftyone import FiftyOne
from .frames import Frames
from .media import Media
from .plugins import Plugins
from .select import Select
from .sort import Sort
from .screenshot import Screenshot
from .tag import Tag
from .tagging import Tagging
from .values import Values
from fiftyone.operators.server import OperatorRoutes

# Starlette routes should not be created here. Please leave as tuple definitions
routes = (
    EmbeddingsRoutes
    + OperatorRoutes
    + [
        ("/aggregate", Aggregate),
        ("/event", Event),
        ("/events", Events),
        ("/fiftyone", FiftyOne),
        ("/frames", Frames),
        ("/media", Media),
        ("/plugins", Plugins),
        ("/select", Select),
        ("/sort", Sort),
        ("/screenshot/{img:str}", Screenshot),
        ("/tag", Tag),
        ("/tagging", Tagging),
        ("/values", Values),
    ]
)



"""
File storage utilities.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from contextlib import contextmanager
from datetime import datetime
import enum
import json
import logging
import multiprocessing.dummy
import ntpath
import os
import posixpath
import re
import six
import shutil
import tempfile

import jsonlines
import yaml

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.utils as fou


logger = logging.getLogger(__name__)


class FileSystem(enum.Enum):
    """Enumeration of the available file systems."""

    LOCAL = "local"


def get_file_system(path):
    """Returns the file system enum for the given path.

    Args:
        path: a path

    Returns:
        a :class:`FileSystem` value
    """
    return FileSystem.LOCAL


def split_prefix(path):
    """Splits the file system prefix from the given path.

    The prefix for local paths is ``""``.

    Example usages::

        import fiftyone.core.storage as fos

        fos.split_prefix("/path/to/file")       # ('', '/path/to/file')
        fos.split_prefix("a/file")              # ('', 'a/file')

    Args:
        path: a path

    Returns:
        a ``(prefix, path)`` tuple
    """
    return "", path


def get_bucket_name(path):
    """Gets the bucket name from the given path.

    The bucket name for local paths is ``""``.

    Example usages::

        import fiftyone.core.storage as fos

        fos.get_bucket_name("/path/to/file")       # ''
        fos.get_bucket_name("a/file")              # ''

    Args:
        path: a path

    Returns:
        the bucket name string
    """
    return ""


def is_local(path):
    """Determines whether the given path is local.

    Args:
        path: a path

    Returns:
        True/False
    """
    return get_file_system(path) == FileSystem.LOCAL


def ensure_local(path):
    """Ensures that the given path is local.

    Args:
        path: a path
    """
    if not is_local(path):
        raise ValueError(
            "The requested operation requires a local path, but found '%s'"
            % path
        )


def normalize_path(path):
    """Normalizes the given path by converting it to an absolute path and
    expanding the user directory, if necessary.

    Args:
        path: a path

    Returns:
        the normalized path
    """
    return os.path.abspath(os.path.expanduser(path))


def make_temp_dir(basedir=None):
    """Makes a temporary directory.

    Args:
        basedir (None): an optional directory in which to create the new
            directory. The default is
            ``fiftyone.config.default_dataset_dir``

    Returns:
        the temporary directory path
    """
    if basedir is None:
        basedir = fo.config.default_dataset_dir

    ensure_dir(basedir)
    return tempfile.mkdtemp(dir=basedir)


class TempDir(object):
    """Context manager that creates and destroys a temporary directory.

    Args:
        basedir (None): an optional directory in which to create the new
            directory. The default is ``fiftyone.config.default_dataset_dir``
    """

    def __init__(self, basedir=None):
        self._basedir = basedir
        self._name = None

    def __enter__(self):
        self._name = make_temp_dir(basedir=self._basedir)
        return self._name

    def __exit__(self, *args):
        delete_dir(self._name)


@contextmanager
def open_file(path, mode="r"):
    """Opens the given file for reading or writing.

    This function *must* be used as a context manager.

    Example usage::

        import fiftyone.core.storage as fos

        with fos.open_file("/tmp/file.txt", "w") as f:
            f.write("Hello, world!")

        with fos.open_file("/tmp/file.txt", "r") as f:
            print(f.read())

    Args:
        path: the path
        mode ("r"): the mode. Supported values are ``("r", "rb", "w", "wb")``
    """
    f = open(path, mode)

    try:
        yield f
    finally:
        f.close()


def read_file(path, binary=False):
    """Reads the file.

    Args:
        path: the filepath
        binary (False): whether to read the file in binary mode

    Returns:
        the file contents
    """
    mode = "rb" if binary else "r"
    with open(path, mode) as f:
        return f.read()


def write_file(str_or_bytes, path):
    """Writes the given string/bytes to a file.

    If a string is provided, it is encoded via ``.encode()``.

    Args:
        str_or_bytes: the string or bytes
        path: the filepath
    """
    ensure_basedir(path)
    with open(path, "wb") as f:
        f.write(_to_bytes(str_or_bytes))


def sep(path):
    """Returns the path separator for the given path.

    For local paths, ``os.path.sep`` is returned.

    For remote paths, ``"/"`` is returned.

    Args:
        path: the filepath

    Returns:
        the path separator
    """
    return os.path.sep


def join(a, *p):
    """Joins the given path components into a single path.

    Args:
        a: the root
        *p: additional path components

    Returns:
        the joined path
    """
    return os.path.join(a, *p)


def isabs(path):
    """Determines whether the given path is absolute.

    Remote paths are always considered absolute.

    Args:
        path: the filepath

    Returns:
        True/False
    """
    return os.path.isabs(path)


def abspath(path):
    """Converts the given path to an absolute path.

    Remote paths are returned unchanged.

    Args:
        path: the filepath

    Returns:
        the absolute path
    """
    return os.path.abspath(path)


def normpath(path):
    """Normalizes the given path by converting all slashes to forward slashes
    on Unix and backslashes on Windows and removing duplicate slashes.

    Use this function when you need a version of ``os.path.normpath`` that
    converts ``\\`` to ``/`` on Unix.

    Args:
        path: a path

    Returns:
        the normalized path
    """
    if os.name == "nt":
        return ntpath.normpath(path)

    return posixpath.normpath(path.replace("\\", "/"))


def exists(path):
    """Determines whether the given file or directory exists.

    Args:
        path: the file or directory path

    Returns:
        True/False
    """
    return os.path.exists(path)


def isfile(path):
    """Determines whether the given file exists.

    Args:
        path: the filepath

    Returns:
        True/False
    """
    return os.path.isfile(path)


def isdir(dirpath):
    """Determines whether the given directory exists.

    Cloud "folders" are deemed to exist only if they are non-empty.

    Args:
        dirpath: the directory path

    Returns:
        True/False
    """
    return os.path.isdir(dirpath)


def make_archive(dirpath, archive_path, cleanup=False):
    """Makes an archive containing the given directory.

    Supported formats include ``.zip``, ``.tar``, ``.tar.gz``, ``.tgz``,
    ``.tar.bz`` and ``.tbz``.

    Args:
        dirpath: the directory to archive
        archive_path: the archive path to write
        cleanup (False): whether to delete the directory after archiving it
    """
    etau.make_archive(dirpath, archive_path)

    if cleanup:
        delete_dir(dirpath)


def extract_archive(archive_path, outdir=None, cleanup=False):
    """Extracts the contents of an archive.

    The following formats are guaranteed to work:
    ``.zip``, ``.tar``, ``.tar.gz``, ``.tgz``, ``.tar.bz``, ``.tbz``.

    If an archive *not* in the above list is found, extraction will be
    attempted via the ``patool`` package, which supports many formats but may
    require that additional system packages be installed.

    Args:
        archive_path: the archive path
        outdir (None): the directory into which to extract the archive. By
            default, the directory containing the archive is used
        cleanup (False): whether to delete the archive after extraction
    """
    if outdir is None:
        outdir = os.path.dirname(archive_path) or "."

    etau.extract_archive(archive_path, outdir=outdir)

    if cleanup:
        delete_file(archive_path)


def ensure_empty_dir(dirpath, cleanup=False):
    """Ensures that the given directory exists and is empty.

    Args:
        dirpath: the directory path
        cleanup (False): whether to delete any existing directory contents

    Raises:
        ValueError: if the directory is not empty and ``cleanup`` is False
    """
    etau.ensure_empty_dir(dirpath, cleanup=cleanup)


def ensure_basedir(path):
    """Makes the base directory of the given path, if necessary.

    Args:
        path: the filepath
    """
    etau.ensure_basedir(path)


def ensure_dir(dirpath):
    """Makes the given directory, if necessary.

    Args:
        dirpath: the directory path
    """
    etau.ensure_dir(dirpath)


def load_json(path_or_str):
    """Loads JSON from the input argument.

    Args:
        path_or_str: the filepath or JSON string

    Returns:
        the loaded JSON
    """
    try:
        return json.loads(path_or_str)
    except ValueError:
        pass

    if os.path.isfile(path_or_str):
        return read_json(path_or_str)

    raise ValueError("Unable to load JSON from '%s'" % path_or_str)


def read_json(path):
    """Reads a JSON file.

    Args:
        path: the filepath

    Returns:
        the JSON data
    """
    try:
        with open(path, "r") as f:
            return json.load(f)
    except ValueError:
        raise ValueError("Unable to parse JSON file '%s'" % path)


def write_json(d, path, pretty_print=False):
    """Writes JSON object to file.

    Args:
        d: JSON data
        path: the filepath
        pretty_print (False): whether to render the JSON in human readable
            format with newlines and indentations
    """
    s = etas.json_to_str(d, pretty_print=pretty_print)
    write_file(s, path)


def load_ndjson(path_or_str):
    """Loads NDJSON from the input argument.

    Args:
        path_or_str: the filepath or NDJSON string

    Returns:
        a list of JSON dicts
    """
    try:
        return etas.load_ndjson(path_or_str)
    except ValueError:
        pass

    if os.path.isfile(path_or_str):
        return read_ndjson(path_or_str)

    raise ValueError("Unable to load NDJSON from '%s'" % path_or_str)


def read_ndjson(path):
    """Reads an NDJSON file.

    Args:
        path: the filepath

    Returns:
        a list of JSON dicts
    """
    with open(path, "r") as f:
        with jsonlines.Reader(f) as r:
            return list(r.iter(skip_empty=True))


def write_ndjson(obj, path):
    """Writes the list of JSON dicts in NDJSON format.

    Args:
        obj: a list of JSON dicts
        path: the filepath
    """
    with open(path, "w") as f:
        with jsonlines.Writer(f) as w:
            w.write_all(obj)


def read_yaml(path):
    """Reads a YAML file.

    Args:
        path: the filepath

    Returns:
        a list of JSON dicts
    """
    with open(path, "r") as f:
        return yaml.safe_load(f)


def write_yaml(obj, path, **kwargs):
    """Writes the object to a YAML file.

    Args:
        obj: a Python object
        path: the filepath
        **kwargs: optional arguments for ``yaml.dump(..., **kwargs)``
    """
    with open(path, "w") as f:
        return yaml.dump(obj, stream=f, **kwargs)


def list_files(
    dirpath,
    abs_paths=False,
    recursive=False,
    include_hidden_files=False,
    return_metadata=False,
    sort=True,
):
    """Lists the files in the given directory.

    If the directory does not exist, an empty list is returned.

    Args:
        dirpath: the path to the directory to list
        abs_paths (False): whether to return the absolute paths to the files
        recursive (False): whether to recursively traverse subdirectories
        include_hidden_files (False): whether to include dot files
        return_metadata (False): whether to return metadata dicts for each file
            instead of filepaths
        sort (True): whether to sort the list of files

    Returns:
        a list of filepaths or metadata dicts
    """
    if not os.path.isdir(dirpath):
        return []

    filepaths = etau.list_files(
        dirpath,
        abs_paths=abs_paths,
        recursive=recursive,
        include_hidden_files=include_hidden_files,
        sort=sort,
    )

    if not return_metadata:
        return filepaths

    metadata = []
    for filepath in filepaths:
        if abs_paths:
            fp = filepath
        else:
            fp = os.path.join(dirpath, filepath)

        m = _get_local_metadata(fp)
        m["filepath"] = filepath
        metadata.append(m)

    return metadata


def _get_local_metadata(filepath):
    s = os.stat(filepath)
    return {
        "name": os.path.basename(filepath),
        "size": s.st_size,
        "last_modified": datetime.fromtimestamp(s.st_mtime),
    }


def list_subdirs(dirpath, abs_paths=False, recursive=False):
    """Lists the subdirectories in the given directory, sorted alphabetically
    and excluding hidden directories.

    Args:
        dirpath: the path to the directory to list
        abs_paths (False): whether to return absolute paths
        recursive (False): whether to recursively traverse subdirectories

    Returns:
        a list of subdirectories
    """
    return etau.list_subdirs(dirpath, abs_paths=abs_paths, recursive=recursive)


def list_buckets(fs, abs_paths=False):
    """Lists the available buckets in the given file system.

    This method returns subdirectories of ``/`` (or the current drive on
    Windows).

    Args:
        fs: a :class:`FileSystem` value
        abs_paths (False): whether to return absolute paths

    Returns:
        a list of buckets
    """
    root = os.path.abspath(os.sep)
    return etau.list_subdirs(root, abs_paths=abs_paths, recursive=False)


def list_available_file_systems():
    """Lists the file systems that are currently available for use with methods
    like :func:`list_files` and :func:`list_buckets`.

    Returns:
        a list of :class:`FileSystem` values
    """
    return [FileSystem.LOCAL]


def get_glob_matches(glob_patt):
    """Returns a list of file paths matching the given glob pattern.

    The matches are returned in sorted order.

    Args:
        glob_patt: a glob pattern like ``/path/to/files-*.jpg``

    Returns:
        a list of file paths
    """
    return etau.get_glob_matches(glob_patt)


def get_glob_root(glob_patt):
    """Finds the root directory of the given glob pattern, i.e., the deepest
    subdirectory that contains no glob characters.

    Args:
        glob_patt: a glob pattern like ``/path/to/files-*.jpg``

    Returns:
        a tuple of:

        -   the root
        -   True/False whether the pattern contains any special characters
    """
    special_chars = "*?[]"

    # Remove escapes around special characters
    replacers = [("[%s]" % s, s) for s in special_chars]
    glob_patt = etau.replace_strings(glob_patt, replacers)

    # @todo optimization: don't split on specials that were previously escaped,
    # as this could cause much more recursive listing than necessary
    split_patt = "|".join(map(re.escape, special_chars))
    root = re.split(split_patt, glob_patt, 1)[0]

    found_special = root != glob_patt
    root = os.path.dirname(root)

    return root, found_special


def copy_file(inpath, outpath):
    """Copies the input file to the output location.

    Args:
        inpath: the input path
        outpath: the output path
    """
    _copy_file(inpath, outpath, cleanup=False)


def copy_files(inpaths, outpaths, skip_failures=False, progress=None):
    """Copies the files to the given locations.

    Args:
        inpaths: a list of input paths
        outpaths: a list of output paths
        skip_failures (False): whether to gracefully continue without raising
            an error if an operation fails
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead
    """
    _copy_files(inpaths, outpaths, skip_failures, progress)


def copy_dir(
    indir, outdir, overwrite=True, skip_failures=False, progress=None
):
    """Copies the input directory to the output directory.

    Args:
        indir: the input directory
        outdir: the output directory
        overwrite (True): whether to delete an existing output directory (True)
            or merge its contents (False)
        skip_failures (False): whether to gracefully continue without raising
            an error if an operation fails
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead
    """
    if overwrite and os.path.isdir(outdir):
        delete_dir(outdir)

    files = list_files(
        indir, include_hidden_files=True, recursive=True, sort=False
    )
    inpaths = [os.path.join(indir, f) for f in files]
    outpaths = [os.path.join(outdir, f) for f in files]
    copy_files(
        inpaths, outpaths, skip_failures=skip_failures, progress=progress
    )


def move_file(inpath, outpath):
    """Moves the given file to a new location.

    Args:
        inpath: the input path
        outpath: the output path
    """
    _copy_file(inpath, outpath, cleanup=True)


def move_files(inpaths, outpaths, skip_failures=False, progress=None):
    """Moves the files to the given locations.

    Args:
        inpaths: a list of input paths
        outpaths: a list of output paths
        skip_failures (False): whether to gracefully continue without raising
            an error if an operation fails
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead
    """
    tasks = [(i, o, skip_failures) for i, o in zip(inpaths, outpaths)]
    if tasks:
        _run(_do_move_file, tasks, progress=progress)


def move_dir(
    indir, outdir, overwrite=True, skip_failures=False, progress=None
):
    """Moves the contents of the given directory into the given output
    directory.

    Args:
        indir: the input directory
        outdir: the output directory
        overwrite (True): whether to delete an existing output directory (True)
            or merge its contents (False)
        skip_failures (False): whether to gracefully continue without raising
            an error if an operation fails
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead
    """
    if overwrite and os.path.isdir(outdir):
        delete_dir(outdir)

    if overwrite:
        etau.ensure_basedir(outdir)
        shutil.move(indir, outdir)


def delete_file(path):
    """Deletes the file at the given path.

    Any empty directories are also recursively deleted from the resulting
    directory tree.

    Args:
        path: the filepath
    """
    _delete_file(path)


def delete_files(paths, skip_failures=False, progress=None):
    """Deletes the files from the given locations.

    Any empty directories are also recursively deleted from the resulting
    directory tree.

    Args:
        paths: a list of paths
        skip_failures (False): whether to gracefully continue without raising
            an error if an operation fails
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead
    """
    tasks = [(p, skip_failures) for p in paths]
    if tasks:
        _run(_do_delete_file, tasks, progress=progress)


def delete_dir(dirpath):
    """Deletes the given directory and recursively deletes any empty
    directories from the resulting directory tree.

    Args:
        dirpath: the directory path
    """
    etau.delete_dir(dirpath)


def run(fcn, tasks, num_workers=None, progress=None):
    """Applies the given function to each element of the given tasks.

    Args:
        fcn: a function that accepts a single argument
        tasks: an iterable of function arguments
        num_workers (None): a suggested number of threads to use
        progress (None): whether to render a progress bar (True/False), use the
            default value ``fiftyone.config.show_progress_bars`` (None), or a
            progress callback function to invoke instead

    Returns:
        the list of function outputs
    """
    num_workers = fou.recommend_thread_pool_workers(num_workers)

    try:
        num_tasks = len(tasks)
    except:
        num_tasks = None

    kwargs = dict(total=num_tasks, iters_str="files", progress=progress)

    if num_workers <= 1:
        with fou.ProgressBar(**kwargs) as pb:
            results = [fcn(task) for task in pb(tasks)]
    else:
        with multiprocessing.dummy.Pool(processes=num_workers) as pool:
            with fou.ProgressBar(**kwargs) as pb:
                results = list(pb(pool.imap(fcn, tasks)))

    return results


def _copy_files(inpaths, outpaths, skip_failures, progress):
    tasks = [(i, o, skip_failures) for i, o in zip(inpaths, outpaths)]
    if tasks:
        _run(_do_copy_file, tasks, progress=progress)


def _run(fcn, tasks, num_workers=None, progress=None):
    num_workers = fou.recommend_thread_pool_workers(num_workers)

    try:
        num_tasks = len(tasks)
    except:
        num_tasks = None

    kwargs = dict(total=num_tasks, iters_str="files", progress=progress)

    if num_workers <= 1:
        with fou.ProgressBar(**kwargs) as pb:
            for task in pb(tasks):
                fcn(task)
    else:
        with multiprocessing.dummy.Pool(processes=num_workers) as pool:
            with fou.ProgressBar(**kwargs) as pb:
                for _ in pb(pool.imap_unordered(fcn, tasks)):
                    pass


def _do_copy_file(arg):
    inpath, outpath, skip_failures = arg

    try:
        _copy_file(inpath, outpath, cleanup=False)
    except Exception as e:
        if not skip_failures:
            raise

        if skip_failures != "ignore":
            logger.warning(e)


def _do_move_file(arg):
    inpath, outpath, skip_failures = arg

    try:
        _copy_file(inpath, outpath, cleanup=True)
    except Exception as e:
        if not skip_failures:
            raise

        if skip_failures != "ignore":
            logger.warning(e)


def _do_delete_file(arg):
    filepath, skip_failures = arg

    try:
        _delete_file(filepath)
    except Exception as e:
        if not skip_failures:
            raise

        if skip_failures != "ignore":
            logger.warning(e)


def _copy_file(inpath, outpath, cleanup=False):
    etau.ensure_basedir(outpath)
    if cleanup:
        shutil.move(inpath, outpath)


def _delete_file(filepath):
    etau.delete_file(filepath)


def _to_bytes(val, encoding="utf-8"):
    b = val.encode(encoding) if isinstance(val, six.text_type) else val
    if not isinstance(b, six.binary_type):
        raise TypeError("Failed to convert %s to bytes" % type(b))

    return b



"""
FiftyOne plugin context.

| Copyright 2017-2023, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import importlib
import logging
import os
import sys
import traceback

import fiftyone as fo
import fiftyone.plugins as fop

from fiftyone.operators.decorators import plugins_cache
from fiftyone.operators.operator import Operator


logger = logging.getLogger(__name__)

INIT_FILENAME = "__init__.py"


@plugins_cache
def build_plugin_contexts(enabled=True):
    """Returns contexts for all available plugins.

    Args:
        enabled (True): whether to include only enabled plugins (True) or only
            disabled plugins (False) or all plugins ("all")

    Returns:
        a list of :class:`PluginContext` instances
    """
    plugin_contexts = []
    for pd in fop.list_plugins(enabled=enabled):
        pctx = PluginContext(pd)
        pctx.register_all()
        plugin_contexts.append(pctx)

    return plugin_contexts


class PluginContext(object):
    """Context that represents a plugin and the Python objects it creates.

    Args:
        plugin_definition: the :class:`fiftyone.plugins.PluginDefinition` for
            the plugin
    """

    def __init__(self, plugin_definition):
        self.plugin_definition = plugin_definition
        self.instances = []
        self.errors = []

    @property
    def name(self):
        """The plugin name."""
        return self.plugin_definition.name

    @property
    def secrets(self):
        """List of keys for required secrets as specified in the plugin definition."""
        return self.plugin_definition.secrets

    def has_errors(self):
        """Determines whether the plugin has errors.

        Returns:
            True/False
        """
        return bool(self.errors)

    def can_register(self, instance):
        """Determines whether the given operator can be registered.

        Args:
            instance: an :class:`fiftyone.operators.operator.Operator`

        Returns:
            True/False
        """
        if not isinstance(instance, Operator):
            return False

        return self.plugin_definition.can_register_operator(instance.name)

    def register(self, cls):
        """Registers the given operator on the plugin.

        .. note::

            Any errors are logged rather than being raised.

        Args:
            cls: an :class:`fiftyone.operators.operator.Operator` class
        """
        try:
            instance = cls()
            if self.can_register(instance):
                instance.plugin_name = self.name
                if self.secrets:
                    instance.add_secrets(self.secrets)
                self.instances.append(instance)
        except:
            logger.warning(
                f"Failed to register operator {cls.__name__} on plugin {self.name}"
            )
            self.errors.append(traceback.format_exc())

    def register_all(self):
        """Registers all operators defined by the plugin on this context.

        .. note::

            Any errors are logged rather than being raised.
        """
        self.dispose_all()

        try:
            module_dir = self.plugin_definition.directory
            module_path = os.path.join(module_dir, INIT_FILENAME)
            if not os.path.isfile(module_path):
                return

            module_name = os.path.relpath(
                module_dir, fo.config.plugins_dir
            ).replace("/", ".")
            spec = importlib.util.spec_from_file_location(
                module_name, module_path
            )
            module = importlib.util.module_from_spec(spec)
            sys.modules[module.__name__] = module
            spec.loader.exec_module(module)
            module.register(self)
        except:
            logger.warning(
                f"Failed to register operators for plugin {self.name}"
            )
            self.errors.append(traceback.format_exc())

    def dispose_all(self):
        """Disposes all operators from this context."""
        self.instances.clear()
        self.errors.clear()
